"""Provides Type structures.
"""
# Author: Pearu Peterson
# Created: February 2019


import re
import copy
import ctypes
import _ctypes
import inspect
from llvmlite import ir
import warnings

from .targetinfo import TargetInfo
from .utils import check_returns_none
import numba as nb
import numpy as np
from numba.core import typing, datamodel, extending, typeconv
from numba.core.imputils import lower_cast


class TypeSystemManager:
    """Manages context specific aliases.

    Usage:

    .. code-block:: python

        with Type.alias(A='Array', bool='bool8', ...):
            # Type.fromstring('A a') will be replaced with Type.fromstring('Array a')

    """

    def alias(self, aliases):
        self.aliases = aliases
        return self

    def __enter__(self):
        self.old_aliases = Type.aliases.copy()
        new_aliases = self.old_aliases.copy()
        new_aliases.update(self.aliases)
        Type.aliases.clear()
        Type.aliases.update(new_aliases)

    def __exit__(self, exc_type, exc_value, traceback):
        Type.aliases.clear()
        Type.aliases.update(self.old_aliases)
        if exc_type is None:
            return True


class TypeParseError(Exception):
    """Failure to parse type definition
    """


def _findparen(s):
    """Find the index of left parenthesis that matches with the one at the
    end of a string.

    Used internally.
    """
    j = s.find(')')
    assert j >= 0, repr((j, s))
    if j == len(s) - 1:
        i = s.find('(')
        if i < 0:
            raise TypeParseError('failed to find lparen index in `%s`' % s)
        return i
    i = s.rfind('(', 0, j)
    if i < 0:
        raise TypeParseError('failed to find lparen index in `%s`' % s)
    t = s[:i] + '_'*(j-i+1) + s[j+1:]
    assert len(t) == len(s), repr((t, s))
    return _findparen(t)


def _commasplit(s):
    """Split a comma-separated items taking into account parenthesis
    and brackets.

    Used internally.
    """
    lst = s.split(',')
    ac = ''
    p1, p2, p3, p4 = 0, 0, 0, 0
    rlst = []
    for i in lst:
        p1 += i.count('(') - i.count(')')
        p2 += i.count('{') - i.count('}')
        p3 += i.count('[') - i.count(']')
        p4 += i.count('<') - i.count('>')
        if p1 == p2 == p3 == p4 == 0:
            rlst.append((ac + ',' + i if ac else i).strip())
            ac = ''
        else:
            ac = ac + ',' + i if ac else i
    if p1 == p2 == p3 == p4 == 0:
        return rlst
    raise TypeParseError('failed to comma-split `%s`' % s)


_booln_match = re.compile(r'\A(boolean|bool|b)(1|8)\Z').match
_charn_match = re.compile(r'\A(char)(32|16|8)(_t|)\Z').match
_intn_match = re.compile(r'\A(signed\s*int|int|i)(\d+)(_t|)\Z').match
_uintn_match = re.compile(r'\A(unsigned\s*int|uint|u)(\d+)(_t|)\Z').match
_floatn_match = re.compile(r'\A(float|f)(16|32|64|128|256)(_t|)\Z').match
_complexn_match = re.compile(
    r'\A(complex|c)(16|32|64|128|256|512)(_t|)\Z').match

_bool_match = re.compile(r'\A(boolean|bool|_Bool|b)\Z').match
_string_match = re.compile(r'\A(string|str)\Z').match

_char_match = re.compile(r'\A(char)\Z').match
_schar_match = re.compile(r'\A(signed\s*char)\Z').match
_uchar_match = re.compile(r'\A(unsigned\s*char|uchar)\Z').match
_byte_match = re.compile(r'\A(signed\s*byte|byte)\Z').match
_ubyte_match = re.compile(r'\A(unsigned\s*byte|ubyte)\Z').match
_wchar_match = re.compile(r'\A(wchar)(_t|)\Z').match

_short_match = re.compile(
    r'\A(signed\s*short\s*int|signed\s*short|short\s*int|short)\Z').match
_ushort_match = re.compile(
    r'\A(unsigned\s*short\s*int|unsigned\s*short|ushort)\Z').match

_int_match = re.compile(r'\A(signed\s*int|signed|int|i)\Z').match
_uint_match = re.compile(r'\A(unsigned\s*int|unsigned|uint|u)\Z').match

_long_match = re.compile(
    r'\A(signed\s*long\s*int|signed\s*long|long\s*int|long|l)\Z').match
_ulong_match = re.compile(
    r'\A(unsigned\s*long\s*int|unsigned\s*long|ulong)\Z').match

_longlong_match = re.compile(
    r'\A(signed\s*long\s*long\s*int|signed\s*long\s*long|long\s*long'
    r'\s*int|long\s*long)\Z').match
_ulonglong_match = re.compile(
    r'\A(unsigned\s*long\s*long\s*int|unsigned\s*long\s*long)\Z'
).match

_size_t_match = re.compile(r'\A(std::|c_|)size_t\Z').match
_ssize_t_match = re.compile(r'\A(std::|c_|)ssize_t\Z').match

_float_match = re.compile(r'\A(float|f)\Z').match
_double_match = re.compile(r'\A(double|d)\Z').match
_longdouble_match = re.compile(r'\A(long\s*double)\Z').match

_complex_match = re.compile(r'\A(complex|c)\Z').match


# For `Type.fromstring('<typespec> <name>')` support:
_type_name_match = re.compile(r'\A(.*)\s(\w+)\Z').match
# bad names are the names of types that can have modifiers such as
# `signed`, `unsigned`, `long`, etc.:
_bad_names_match = re.compile(r'\A(char|byte|short|int|long|double)\Z').match

# For the custom type support:
_custom_type_name_params_match = re.compile(r'\A(\w+)\s*[<](.*)[>]\Z').match


class Complex64(ctypes.Structure):
    _fields_ = [("real", ctypes.c_float), ("imag", ctypes.c_float)]

    @classmethod
    def from_param(cls, obj):
        if isinstance(obj, complex):
            return cls(obj.real, obj.imag)
        if isinstance(obj, (int, float)):
            return cls(obj.real, 0.0)
        raise NotImplementedError(repr(type(obj)))


class Complex128(ctypes.Structure):
    _fields_ = [("real", ctypes.c_double), ("imag", ctypes.c_double)]

    @classmethod
    def from_param(cls, obj):
        if isinstance(obj, complex):
            return cls(obj.real, obj.imag)
        if isinstance(obj, (int, float)):
            return cls(obj.real, 0.0)
        raise NotImplementedError(repr(type(obj)))

    def topython(self):
        return complex(self.real, self.imag)


# Initialize type maps
_ctypes_imap = {
    ctypes.c_void_p: 'void*', None: 'void', ctypes.c_bool: 'bool',
    ctypes.c_char_p: 'char%s*' % (ctypes.sizeof(ctypes.c_char()) * 8),
    ctypes.c_wchar_p: 'char%s*' % (ctypes.sizeof(ctypes.c_wchar()) * 8),
}
_ctypes_char_map = {}
_ctypes_int_map = {}
_ctypes_uint_map = {}
_ctypes_float_map = {}
_ctypes_complex_map = {}
for _k, _m, _lst in [
        ('char', _ctypes_char_map, ['c_char', 'c_wchar']),
        ('int', _ctypes_int_map,
         ['c_int8', 'c_int16', 'c_int32', 'c_int64', 'c_int',
          'c_long', 'c_longlong', 'c_byte', 'c_short', 'c_ssize_t']),
        ('uint', _ctypes_uint_map,
         ['c_uint8', 'c_uint16', 'c_uint32', 'c_uint64', 'c_uint',
          'c_ulong', 'c_ulonglong', 'c_ubyte', 'c_ushort', 'c_size_t']),
        ('float', _ctypes_float_map,
         ['c_float', 'c_double', 'c_longdouble']),
        ('complex', _ctypes_complex_map,
         [Complex64, Complex128])
]:
    for _n in _lst:
        if isinstance(_n, str):
            _t = getattr(ctypes, _n, None)
        else:
            _t = _n
        if _t is not None:
            _b = ctypes.sizeof(_t) * 8
            if _b not in _m:
                _m[_b] = _t
            _ctypes_imap[_t] = _k + str(_b)

_numba_imap = {nb.void: 'void', nb.boolean: 'bool'}
_numba_char_map = {}
_numba_bool_map = {}
_numba_int_map = {}
_numba_uint_map = {}
_numba_float_map = {}
_numba_complex_map = {}
for _k, _m, _lst in [
        ('int', _numba_int_map,
            ['int8', 'int16', 'int32', 'int64', 'intc', 'int_', 'intp',
             'long_', 'longlong', 'short', 'char']),
        ('uint', _numba_uint_map,
            ['uint8', 'uint16', 'uint32', 'uint64', 'uintc', 'uint',
             'uintp', 'ulong', 'ulonglong', 'ushort']),
        ('float', _numba_float_map,
            ['float32', 'float64', 'float_', 'double']),
        ('complex', _numba_complex_map, ['complex64', 'complex128']),
]:
    for _n in _lst:
        _t = getattr(nb, _n, None)
        if _t is not None:
            _b = _t.bitwidth
            if _b not in _m:
                _m[_b] = _t
            _numba_imap[_t] = _k + str(_b)

# numpy mapping
_numpy_imap = {}
for v in set(np.typeDict.values()):
    name = np.dtype(v).name
    _numpy_imap[v] = name

# python_imap values must be processed with Type.fromstring
_python_imap = {int: 'int64', float: 'float64', complex: 'complex128',
                str: 'string', bytes: 'char*'}

# Data for the mangling algorithm, see mangle/demangle methods.
#
_mangling_suffices = '_VW'
_mangling_prefixes = 'PKaAMrR'
_mangling_map = dict(
    void='v', bool='b',
    char8='c', char16='z', char32='w',
    int8='B', int16='s', int32='i', int64='l', int128='q',
    uint8='U', uint16='S', uint32='I', uint64='L', uint128='Q',
    float16='h', float32='f', float64='d', float128='x',
    complex32='H', complex64='F', complex128='D', complex256='X',
    string='t',
)
_mangling_imap = {}
for _k, _v in _mangling_map.items():
    assert _v not in _mangling_imap, repr((_k, _v))
    assert len(_v) == 1, repr((_k, _v))
    _mangling_imap[_v] = _k
# make sure that mangling keys will not conflict with mangling
# operators:
_i = set(_mangling_imap).intersection(_mangling_suffices+_mangling_prefixes)
assert not _i, repr(_i)


class MetaType(type):

    custom_types = dict()
    aliases = dict()

    # ctypes generated types need to be cached to be usable
    ctypes_types = dict()

    def __new__(cls, name, bases, dct):
        cls = super().__new__(cls, name, bases, dct)
        if name != 'Type':
            cls.custom_types[name] = cls
        return cls

    def alias(cls, **aliases):
        """
        Define type aliases context. For instance,

          with Type.alias(myint='int64'):
              ...

        will ensure under the with block we have

          Type.fromstring('myint') == Type.fromstring('int64')
        """
        return TypeSystemManager().alias(aliases)


class Type(tuple, metaclass=MetaType):
    """Represents a type.

    There are six kinds of a types:

      ==========    ==============================    ==============================
      Type          Description                       Internal structure
      ==========    ==============================    ==============================
      void          a "no type"                       Type()
      atomic        e.g. ``int32``                    Type(<str>,)
      pointer       e.g. ``int32*``                   Type(<Type instance>, '*')
      struct        e.g. ``{int32, int32}``           Type(<Type instances>, <Type instances>, ...)
      function      e.g. ``int32(int32, int32)``      Type(<Type instance>, ..., name='')
      custom        e.g. ``MyClass<int32, int32>``    Type((<object>,))
      undefined     e.g. ``fromcallable(foo)``        Type(None)
      ==========    ==============================    ==============================

    Atomic types are types with names (Type contains a single
    string). All other types (except "no type") are certain
    implementations of atomic types.

    The name content of an atomic type is arbitrary but it cannot be
    empty. For instance, Type('a') and Type('a long name') are atomic
    types.

    Parsing types from a string is not fixed to any type system, the
    names of types can be arbitrary.  However, converting the Type
    instances to concrete types such as provided in numpy or numba,
    the following atomic types are defined (the first name corresponds
    to normalized name)::

      no type:                    void, none
      bool:                       bool, boolean, _Bool, b
      1-bit bool:                 bool1, boolean1, b1
      8-bit bool:                 bool8, boolean8, b8
      8-bit char:                 char8, char
      16-bit char:                char16
      32-bit char:                char32, wchar
      8-bit signed integer:       int8, i8, byte, signed char
      16-bit signed integer:      int16, i16, int16_t
      32-bit signed integer:      int32, i32, int32_t
      64-bit signed integer:      int64, i64, int64_t
      128-bit signed integer:     int128, i128, int128_t
      8-bit unsigned integer:     uint8, u8, ubyte, unsigned char
      16-bit unsigned integer:    uint16, u16, uint16_t
      32-bit unsigned integer:    uint32, u32, uint32_t
      64-bit unsigned integer:    uint64, u64, uint64_t
      128-bit unsigned integer:   uint128, u128, uint64_t
      16-bit float:               float16, f16
      32-bit float:               float32, f32, float
      64-bit float:               float64, f64, double
      128-bit float:              float128, f128, long double
      32-bit complex:             complex32, c32, complex
      64-bit complex:             complex64, c64
      128-bit complex:            complex128, c128
      256-bit complex:            complex256, c256
      string:                     string, str

    with the following extensions::

      N-bit signed integer:   int<N>, i<N>       for instance: int5, i31
      N-bit unsigned integer: uint<N>, u<N>
      N-bit float:            float<N>
      N-bit complex:          complex<N>

    Also ``byte, short, int, long, long long, signed int, size_t,
    ssize_t``, etc are supported but their normalized names are system
    dependent.

    Custom types
    ------------

    The typesystem supports processing custom types that are usually
    named structures or C++ template specifications, but not only.

    Custom type can have arbitrary number of (hashable) parameters.  A
    custom type is concrete when the parameters with type Type are
    concrete.

    Custom types must implement tonumba method if needed. For that,
    one must derive MyClass from Type.

    Internally, a custom type has two possible representations: if
    MyClass is derived from Type then `MyClass<a, b>` is represented
    as `MyClass(('a', 'b'))`, otherwise, it is represented as
    `Type(('MyClass', 'a', 'b'))`. If the parameters of a custom types
    need to be Type instances, use `preprocess_args` for required
    conversion of arguments.
    """

    _mangling = None

    def __new__(cls, *args, **params):
        args = cls.preprocess_args(args)
        obj = tuple.__new__(cls, args)
        obj._params = params
        if not obj._is_ok:
            if obj._is_function:
                raise ValueError('cannot create named function type from `%s`' % (args,))
            raise ValueError(
                'attempt to create an invalid Type object from `%s`' % (args,))
        return obj.postprocess_type()

    def copy(self, cls=None):
        """Return a copy of type.
        """
        if cls is None:
            cls = type(self)
        return cls(*self, **copy.deepcopy(self._params))

    @classmethod
    def preprocess_args(cls, args):
        """Preprocess the arguments of Type constructor.

        Overloading this classmethod may be useful for custom types
        for processing its parameters.
        """
        return args

    def postprocess_type(self):
        """Postprocess Type construction. Must return Type instance.
        """
        return self

    def annotation(self, **annotations):
        """Set and get annotations.
        """
        annotation = self._params.get('annotation')
        if annotation is None:
            annotation = self._params['annotation'] = {}
        annotation.update(annotations)
        return annotation

    def __or__(self, other):
        """Apply annotations to a copy of type instance.
        """
        self.annotation()  # ensures annotation key in _params
        params = self._params.copy()
        annotation = params['annotation'] = params['annotation'].copy()
        if isinstance(other, str):
            if other:
                annotation[other] = ''
        elif isinstance(other, dict):
            annotation.update(other)
        return type(self)(*self, **params)

    def inherit_annotations(self, other):
        if isinstance(other, Type):
            self.annotation().update(other.annotation())
            for a, b in zip(self, other):
                if b is None:
                    continue
                if isinstance(a, str):
                    pass
                elif isinstance(a, Type):
                    a.inherit_annotations(b)
                elif isinstance(a, tuple):
                    for x, y in zip(a, b):
                        if isinstance(x, str):
                            pass
                        else:
                            x.inherit_annotations(y)
                else:
                    raise NotImplementedError('inherit_annotations: %s'
                                              % ((a, type(a)),))

    def params(self, other=None, **params):
        """In-place update of parameters from other or/and dictionary and return self.
        """
        if other is not None:
            return self.params(None, **other._params).params(None, **params)
        for k, v in params.items():
            if k == 'annotation':
                self.annotation().update(v)
            else:
                orig_v = self._params.get(k)
                if orig_v is not None and orig_v != v:
                    warnings.warn(
                        f'{type(self).__name__}.params: overwriting '
                        f'existing parameter {k} with {v} (original value is {orig_v})')
                self._params[k] = v
        return self

    def set_mangling(self, mangling):
        """Set mangling string of the type.
        """
        self._mangling = mangling

    def mangling(self):
        if self._mangling is None:
            self._mangling = self.mangle()
        return self._mangling

    @property
    def is_void(self):
        return len(self) == 0

    @property
    def is_atomic(self):
        return len(self) == 1 and isinstance(self[0], str)

    @property
    def is_undefined(self):
        return len(self) == 1 and self[0] is None

    @property
    def is_int(self):
        return self.is_atomic and self[0].startswith('int')

    @property
    def is_uint(self):
        return self.is_atomic and self[0].startswith('uint')

    @property
    def is_float(self):
        return self.is_atomic and self[0].startswith('float')

    @property
    def is_complex(self):
        return self.is_atomic and self[0].startswith('complex')

    @property
    def is_string(self):
        return self.is_atomic and self[0] == 'string'

    @property
    def is_bool(self):
        return self.is_atomic and self[0].startswith('bool')

    @property
    def is_char(self):
        return self.is_atomic and self[0].startswith('char')

    @property
    def is_aggregate(self):
        # ref: https://llvm.org/docs/LangRef.html#aggregate-types
        return self.is_struct

    @property
    def is_pointer(self):
        return len(self) == 2 and isinstance(self[0], Type) \
            and isinstance(self[1], str) and self[1] == '*'

    @property
    def is_struct(self):
        return len(self) > 0 and all(isinstance(s, Type) for s in self)

    @property
    def _is_function(self):
        return len(self) == 2 and isinstance(self[0], Type) and \
            isinstance(self[1], tuple) and not isinstance(self[1], Type) \
            and all([isinstance(p, Type) for p in self[1]])

    @property
    def is_function(self):
        return self._is_function and 'name' in self._params

    @property
    def is_custom(self):
        return len(self) == 1 and not isinstance(self[0], Type) and isinstance(self[0], tuple)

    @property
    def is_signed(self):
        # https://en.cppreference.com/w/cpp/types/numeric_limits/is_signed
        return self.is_atomic and (self.is_char or self.is_int or self.is_float)

    @property
    def is_unsigned(self):
        # https://en.cppreference.com/w/cpp/types/numeric_limits/is_signed
        return self.is_atomic and not self.is_signed

    @property
    def is_complete(self):
        """Return True when the Type instance does not contain unknown types.
        """
        if self.is_atomic:
            return not self[0].startswith('<type of')
        elif self.is_pointer:
            return self[0].is_complete
        elif self.is_struct:
            for m in self:
                if not m.is_complete:
                    return False
        elif self.is_function:
            if not self[0].is_complete:
                return False
            for a in self[1]:
                if not a.is_complete:
                    return False
        elif self.is_void:
            pass
        elif self.is_custom:
            for a in self[0]:
                if isinstance(a, Type) and not a.is_complete:
                    return False
        elif self.is_undefined:
            return False
        else:
            raise NotImplementedError(repr(self))
        return True

    @property
    def is_concrete(self):
        """Return True when the Type instance is concrete.
        """
        if self.is_atomic:
            return (self.is_int or self.is_uint or self.is_float
                    or self.is_complex or self.is_bool
                    or self.is_string or self.is_char)
        elif self.is_pointer:
            return self[0].is_concrete
        elif self.is_struct:
            for m in self:
                if not m.is_concrete:
                    return False
        elif self.is_function:
            if not self[0].is_concrete:
                return False
            for a in self[1]:
                if not a.is_concrete:
                    return False
        elif self.is_void:
            pass
        elif self.is_custom:
            for a in self[0]:
                if isinstance(a, Type) and not a.is_concrete:
                    return False
        elif self.is_undefined:
            return False
        else:
            raise NotImplementedError(repr(self))
        return True

    @property
    def _is_ok(self):
        return self.is_void or self.is_atomic or self.is_pointer \
            or self.is_struct \
            or (self.is_function and len(self[1]) > 0) \
            or self.is_custom or self.is_undefined

    def __repr__(self):
        d = {}
        for k, v in self._params.items():
            if v:
                d[k] = v
        if d:
            return '%s%s|%s' % (type(self).__name__, tuple.__repr__(self), d)
        return '%s%s' % (type(self).__name__, tuple.__repr__(self))

    @property
    def consumes_nargs(self):
        """Return the number of arguments that the given type consumes.
        """
        return len(self.as_consumed_args)

    @property
    def as_consumed_args(self):
        """Return the argument types that the given type will consume.
        """
        return [self]

    @property
    def arity(self):
        """Return the arity of the function type.

        Some function argument types may consume several function
        arguments (e.g., data pointer and data size). The arity of
        function type is the number of functon arguments that are
        consumed when constructing a call.

        Arguments with default value are ignored.
        """
        assert self.is_function
        arity = 0
        for t in self[1]:
            if t.annotation().get('default'):
                break
            arity += t.consumes_nargs
        return arity

    @property
    def argument_types(self):
        """Return the list of consumed argument types.
        """
        assert self.is_function
        atypes = []
        for t in self[1]:
            atypes.extend(t.as_consumed_args)
        return atypes

    @property
    def name(self):
        """Return declarator name.

        Function types always define a name: when not specified then
        the name value will be an empty string.

        For other types the name is optional: when not specified then
        the name value will be None.
        """
        return self._params.get('name')

    def get_field_position(self, name):
        """Return the index of a structure member with name.

        Returns None when no match with member names is found.
        """
        assert self.is_struct
        for i, m in enumerate(self):
            if m.name == name:
                return i

    def __str__(self):
        if self._is_ok:
            return self.tostring()
        return tuple.__str__(self)

    def tostring(self, use_typename=False, use_annotation=True, use_name=True,
                 use_annotation_name=False, _skip_annotation=False):
        """Return string representation of a type.
        """
        options = dict(use_typename=use_typename, use_annotation=use_annotation,
                       use_name=use_name,
                       use_annotation_name=use_annotation_name)
        annotation = self.annotation()
        if use_annotation and not _skip_annotation:
            s = self.tostring(_skip_annotation=True, **options)
            for name, value in annotation.items():
                if (
                        use_annotation_name
                        and use_name
                        and name == 'name'
                        and self._params.get('name', value) == value):
                    continue
                if value:
                    s = '%s | %s=%s' % (s, name, value)
                else:
                    s = '%s | %s' % (s, name)
            return s

        if self.is_void:
            return 'void'

        name = self._params.get('name') if use_name else None
        if use_annotation_name and use_name:
            annot_name = annotation.get('name')
            if annot_name is not None and name is None:
                name = annot_name

        if self.is_function:
            if use_typename:
                typename = self._params.get('typename')
                if typename is not None:
                    return typename
            name = ' ' + name if name else ''
            return (self[0].tostring(**options)
                    + name + '(' + ', '.join(a.tostring(**options) for a in self[1]) + ')')

        if name is not None:
            suffix = ' ' + name
        else:
            suffix = ''
        if use_typename:
            typename = self._params.get('typename')
            if typename is not None:
                return typename + suffix
        if self.is_atomic:
            return self[0] + suffix
        if self.is_pointer:
            return self[0].tostring(**options) + '*' + suffix
        if self.is_struct:
            clsname = self._params.get('clsname')
            if clsname is not None:
                return clsname + '<' + ', '.join(
                    [t.tostring(**options)
                     for t in self]) + '>' + suffix
            return '{' + ', '.join([t.tostring(**options) for t in self]) + '}' + suffix

        if self.is_custom:
            params = self[0]
            if type(self) is Type:
                name = params[0]
                params = params[1:]
            else:
                name = type(self).__name__
            name = self._params.get('shorttypename', name)
            new_params = []
            for a in params:
                if isinstance(a, Type):
                    s = a.tostring(**options)
                else:
                    s = str(a)
                new_params.append(s)
            return (name + '<' + ', '.join(new_params) + '>' + suffix)
        raise NotImplementedError(repr(self))

    def toprototype(self):
        if self.is_void:
            return 'void'
        typename = self._params.get('typename')
        if typename is not None:
            return typename
        if self.is_atomic:
            s = self[0]
            if self.is_int or self.is_uint:
                return s + '_t'
            if self.is_float:
                bits = self.bits
                if bits == 32:
                    return 'float'
                elif bits == 64:
                    return 'double'
            return s
        if self.is_pointer:
            return self[0].toprototype() + '*'
        if self.is_struct:
            return '{' + ', '.join([t.toprototype() for t in self]) + '}'
        if self.is_function:
            return self[0].toprototype() + '(' + ', '.join(
                a.toprototype() for a in self[1]) + ')'
        if self.is_custom:
            raise NotImplementedError(f'{type(self).__name__}.toprototype()')
        raise NotImplementedError(repr(self))

    def tonumba(self, bool_is_int8=None):
        """Convert Type instance to numba type object.

        Parameters
        ----------
        bool_is_int8: {bool, None}

          If true, boolean data and values are mapped to LLVM `i8`,
          otherwise to `i1`. Note that numba boolean maps data to `i8`
          and value to `i1`. To get numba convention, specify
          `bool_is_int8` as `None`.

        """
        numba_type = self._params.get('tonumba')
        if numba_type is not None:
            return numba_type
        if self.is_void:
            return nb.void
        if self.is_int:
            return _numba_int_map.get(int(self[0][3:]))
        if self.is_uint:
            return _numba_uint_map.get(int(self[0][4:]))
        if self.is_float:
            return _numba_float_map.get(int(self[0][5:]))
        if self.is_complex:
            return _numba_complex_map.get(int(self[0][7:]))
        if self.is_bool:
            if bool_is_int8 is None:
                return _numba_bool_map.get(self.bits, nb.boolean)
            return boolean8 if bool_is_int8 else boolean1
        if self.is_pointer:
            if self[0].is_void:
                return nb.types.voidptr
            if self[0].is_struct:
                ptr_type = self._params.get('NumbaType', structure_type.StructureNumbaPointerType)
            else:
                ptr_type = self._params.get('NumbaType', nb.types.CPointer)
            return ptr_type(self[0].tonumba(bool_is_int8=bool_is_int8))
        if self.is_struct:
            struct_name = self._params.get('name')
            if struct_name is None:
                struct_name = 'STRUCT'+self.mangling()
            members = []
            for i, member in enumerate(self):
                name = member._params.get('name', '_%s' % (i+1))
                members.append((name,
                                member.tonumba(bool_is_int8=bool_is_int8)))
            return structure_type.make_numba_struct(struct_name, members, origin=self)
        if self.is_function:
            rtype = self[0].tonumba(bool_is_int8=bool_is_int8)
            atypes = [t.tonumba(bool_is_int8=bool_is_int8)
                      for t in self[1] if not t.is_void]
            return rtype(*atypes)
        if self.is_string:
            return nb.types.string
        if self.is_char:
            # in numba, char==int8
            return _numba_int_map.get(int(self[0][4:]))
        if self.is_atomic:
            return nb.types.Type(self[0])
        if self.is_custom:
            return self.__typesystem_type__.tonumba(bool_is_int8=bool_is_int8)
        raise NotImplementedError(repr(self))

    def toctypes(self):
        """Convert Type instance to ctypes type object.
        """
        if self.is_void:
            return None
        if self.is_int:
            return _ctypes_int_map[int(self[0][3:])]
        if self.is_uint:
            return _ctypes_uint_map[int(self[0][4:])]
        if self.is_float:
            return _ctypes_float_map[int(self[0][5:])]
        if self.is_complex:
            return _ctypes_complex_map[int(self[0][7:])]
        if self.is_bool:
            return ctypes.c_bool
        if self.is_char:
            return _ctypes_char_map[int(self[0][4:])]
        if self.is_pointer:
            if self[0].is_void:
                return ctypes.c_void_p
            if self[0].is_char:
                return getattr(ctypes,
                               _ctypes_char_map.get(
                                   int(self[0][0][4:])).__name__ + '_p')
            return ctypes.POINTER(self[0].toctypes())
        if self.is_struct:
            ctypes_type_name = 'rbc_typesystem_struct_%s' % (self.mangle())
            typ = type(self).ctypes_types.get(ctypes_type_name)
            if typ is None:
                fields = [((t.name if t.name else 'f%s' % i), t.toctypes())
                          for i, t in enumerate(self)]
                typ = type(ctypes_type_name,
                           (ctypes.Structure, ),
                           dict(_fields_=fields))
                type(self).ctypes_types[ctypes_type_name] = typ
            return typ
        if self.is_function:
            ctypes_type_name = 'rbc_typesystem_function_%s' % (self.mangle())
            typ = type(self).ctypes_types.get(ctypes_type_name)
            if typ is None:
                rtype = self[0].toctypes()
                atypes = []
                for t in self[1]:
                    if t.is_struct:
                        # LLVM struct argument (as an aggregate type)
                        # is mapped to struct member arguments:
                        atypes.extend([m.toctypes() for m in t])
                    elif t.is_void:
                        pass
                    else:
                        atypes.append(t.toctypes())
                typ = ctypes.CFUNCTYPE(rtype, *atypes)
                type(self).ctypes_types[ctypes_type_name] = typ
            return typ
        if self.is_string:
            return ctypes.c_wchar_p
        if self.is_custom:
            raise NotImplementedError(f'{type(self).__name__}.toctypes()|self={self}')
        raise NotImplementedError(repr((self, self.is_string)))

    def tollvmir(self, bool_is_int8=None):
        if self.is_int:
            return ir.IntType(self.bits)
        if self.is_float:
            return {32: ir.FloatType, 64: ir.DoubleType}[self.bits]()
        if self.is_bool:
            if bool_is_int8:
                return ir.IntType(8)
            return ir.IntType(self.bits)
        if self.is_pointer:
            if self[0].is_void:
                # mapping void* to int8*
                return ir.IntType(8).as_pointer()
            return self[0].tollvmir(bool_is_int8=bool_is_int8).as_pointer()
        if self.is_void:
            # Used only as the return type of a function without a return value.
            # TODO: ensure that void is used only as function return type or a pointer dtype.
            return ir.VoidType()
        if self.is_struct:
            return ir.LiteralStructType([m.tollvmir(bool_is_int8=bool_is_int8) for m in self])
        raise NotImplementedError(f'{type(self).__name__}.tollvmir()|self={self}')

    @classmethod
    def _fromstring(cls, s):
        s = s.strip()
        if len(s) > 1 and s.endswith('*'):       # pointer
            return cls(cls._fromstring(s[:-1]), '*')
        if s.endswith('}'):       # struct
            if not s.startswith('{'):
                raise TypeParseError(
                    'mismatching curly parenthesis in `%s`' % (s))
            return cls(*map(cls._fromstring,
                            _commasplit(s[1:-1].strip())))
        if s.endswith(')'):       # function
            i = _findparen(s)
            if i < 0:
                raise TypeParseError('mismatching parenthesis in `%s`' % (s))
            atypes = tuple(map(cls._fromstring,
                               _commasplit(s[i+1:-1].strip())))
            r = s[:i].strip()
            if r.endswith(')'):
                j = _findparen(r)
                if j < 0:
                    raise TypeParseError('mismatching parenthesis in `%s`' % (r))
                rtype = cls._fromstring(r[:j])
                d = r[j+1:-1].strip()
                if d.startswith('*'):
                    while d.startswith('*'):
                        d = d[1:].lstrip()
                    if d.endswith(')'):
                        k = _findparen(d)
                        name = d[:k]
                        rtype = cls(rtype, atypes, name='')
                        atypes = tuple(map(cls._fromstring,
                                           _commasplit(d[k+1:-1].strip())))
                        return cls(rtype, atypes, name=name)
                    name = d
                    return cls(rtype, atypes, name=name)
            rtype = cls._fromstring(r)
            if rtype.is_function:
                name = rtype._params['name']
                rtype._params['name'] = ''
            else:
                name = rtype._params.pop('name', '')
            return cls(rtype, atypes, name=name)

        i_bar, i_gt = s.find('|'), s.find('>')  # TODO: will need a better parser
        if '|' in s and (i_gt == -1 or i_bar > i_gt):
            s, a = s.rsplit('|', 1)
            t = cls._fromstring(s.rstrip())
            if '=' in a:
                n, v = a.split('=', 1)
            else:
                n, v = a, ''
            n = n.strip()
            v = v.strip()
            if n or v:
                t.annotation(**{n: v})
            return t
        if s.endswith('>') and not s.startswith('<'):  # custom
            i = s.index('<')
            name = s[:i]
            params = _commasplit(s[i+1:-1].strip())
            params = tuple(map(cls._fromstring, params)) if params else ()
            name = cls.aliases.get(name, name)
            if name in cls.custom_types:
                cls = cls.custom_types[name]
                r = cls(params)
            else:
                r = cls((name,) + params)
            return r
        if s == 'void' or s == 'none' or not s:  # void
            return cls()
        m = _type_name_match(s)
        if m is not None:
            name = m.group(2)
            if not _bad_names_match(name):
                # `<typespec> <name>`
                t = cls._fromstring(m.group(1))
                t._params['name'] = name
                return t
        # atomic
        if s in cls.custom_types:
            return cls.custom_types[s](())
        return cls(s)

    @classmethod
    def fromstring(cls, s):
        """Return new Type instance from a string.

        Parameters
        ----------
        s : str
        """
        try:
            return cls._fromstring(s)._normalize()
        except TypeParseError as msg:
            raise ValueError('failed to parse `%s`: %s' % (s, msg))

    @classmethod
    def fromnumpy(cls, t):
        """Return new Type instance from numpy type object.
        """
        n = _numpy_imap.get(t)
        if n is not None:
            return cls.fromstring(n)

        raise NotImplementedError(repr(t))

    @classmethod
    def fromnumba(cls, t):
        """Return new Type instance from numba type object.
        """
        if hasattr(t, "__typesystem_type__") or hasattr(type(t), "__typesystem_type__"):
            return t.__typesystem_type__
        n = _numba_imap.get(t)
        if n is not None:
            return cls.fromstring(n)
        if isinstance(t, typing.templates.Signature):
            atypes = (cls.fromnumba(a) for a in t.args)
            rtype = cls.fromnumba(t.return_type)
            return cls(rtype, tuple(atypes) or (Type(),), name='')
        if isinstance(t, nb.types.misc.CPointer):
            return cls(cls.fromnumba(t.dtype), '*')
        if isinstance(t, nb.types.NumberClass):
            return cls.fromnumba(t.instance_type)
        if isinstance(t, nb.types.Boolean):
            # boolean1 and boolean8 map both to bool
            return cls.fromstring('bool')
        if isinstance(t, nb.types.misc.RawPointer):
            if t == nb.types.voidptr:
                return cls(cls(), '*')

        raise NotImplementedError(repr((t, type(t).__bases__)))

    @classmethod
    def fromctypes(cls, t):
        """Return new Type instance from ctypes type object.
        """
        n = _ctypes_imap.get(t)
        if n is not None:
            return cls.fromstring(n)
        if issubclass(t, ctypes.Structure):
            return cls(*(cls.fromctypes(_t)
                         for _f, _t in t._fields_))
        if issubclass(t, ctypes._Pointer):
            return cls(cls.fromctypes(t._type_), '*')
        if issubclass(t, ctypes._CFuncPtr):
            atypes = tuple(cls.fromctypes(a) for a in t._argtypes_)
            return cls(cls.fromctypes(t._restype_), atypes or (Type(),), name='')
        raise NotImplementedError(repr(t))

    @classmethod
    def fromcallable(cls, func):
        """Return new Type instance from a callable object.

        The callable object must use annotations for specifying the
        types of arguments and return value.
        """
        if not hasattr(func, '__name__'):
            raise ValueError(
                'constructing Type instance from a callable without `__name__`'
                f' is not supported, got {func}|{type(func).__bases__}')
        if func.__name__ == '<lambda>':
            # lambda function cannot carry annotations, hence:
            raise ValueError('constructing Type instance from '
                             'a lambda function is not supported')
        sig = get_signature(func)
        annot = sig.return_annotation
        if isinstance(annot, dict):
            rtype = cls() | annot  # void
        elif annot == sig.empty:
            if check_returns_none(func):
                rtype = cls()  # void
            else:
                rtype = cls(None)   # cannot deterimine return value type
        else:
            rtype = cls.fromobject(annot)
        atypes = []
        for n, param in sig.parameters.items():
            annot = param.annotation
            if param.kind not in [inspect.Parameter.POSITIONAL_OR_KEYWORD,
                                  inspect.Parameter.POSITIONAL_ONLY]:
                raise ValueError(
                    'callable argument kind must be positional,'
                    ' `%s` has kind %s' % (param, param.kind))
            if param.default != sig.empty:
                if annot == sig.empty:
                    annot = {}
                annot['default'] = param.default
            if isinstance(annot, dict):
                atypes.append(cls('<type of %s>' % n) | annot)
            elif annot == sig.empty:
                atypes.append(cls('<type of %s>' % n))
            else:
                atypes.append(cls.fromobject(annot))

        return cls(rtype, tuple(atypes) or (Type(),), name=func.__name__)

    @classmethod
    def fromvalue(cls, obj):
        """Return Type instance that corresponds to given value.
        """
        for mapping in [_python_imap, _numpy_imap]:
            n = mapping.get(type(obj))
            if n is not None:
                return cls.fromstring(n)
        if isinstance(obj, _ctypes._Pointer):
            return cls.fromctypes(obj._type_).pointer()
        if isinstance(obj, ctypes.c_void_p):
            return cls(cls(), '*')
        if hasattr(obj, '__typesystem_type__'):
            return cls.fromobject(obj.__typesystem_type__)

        raise NotImplementedError('%s.fromvalue(%r|%s)'
                                  % (cls.__name__, obj, type(obj)))

    @classmethod
    def fromobject(cls, obj):
        """Return new Type instance from any object.

        Parameters
        ----------
        obj : object
        """
        if isinstance(obj, cls):
            return obj
        if isinstance(obj, str):
            return cls.fromstring(obj)
        n = _python_imap.get(obj)
        if n is not None:
            return cls.fromstring(n)
        if hasattr(obj, "__typesystem_type__") or hasattr(type(obj), "__typesystem_type__"):
            return obj.__typesystem_type__
        if hasattr(obj, '__module__'):
            if obj.__module__.startswith('numba'):
                return cls.fromnumba(obj)
            if obj.__module__.startswith('ctypes'):
                return cls.fromctypes(obj)
            if obj.__module__.startswith('numpy'):
                return cls.fromnumpy(obj)
        if inspect.isclass(obj):
            if obj is int:
                return cls('int64')
            return cls.fromstring(obj.__name__)
        if callable(obj):
            return cls.fromcallable(obj)
        raise NotImplementedError(repr((type(obj))))

    def _normalize(self):
        """Return new Type instance with atomic types normalized.
        """
        params = self._params
        if self.is_void:
            return self
        if self.is_atomic:
            s = self[0]
            a = type(self).aliases.get(s)
            if a is not None:
                return Type.fromobject(a).params(self)._normalize()
            m = _string_match(s)
            if m is not None:
                return Type('string', **params)
            for match, ntype in [
                    (_booln_match, 'bool'),
                    (_charn_match, 'char'),
                    (_intn_match, 'int'),
                    (_uintn_match, 'uint'),
                    (_floatn_match, 'float'),
                    (_complexn_match, 'complex'),
            ]:
                m = match(s)
                if m is not None:
                    bits = m.group(2)
                    return Type(ntype + bits, **params)
            if s.endswith('[]'):
                return Type.fromstring(f'Array<{s[:-2]}>')._normalize()
            m = _bool_match(s)
            if m is not None:
                return Type('bool', **params)
            if _char_match(s):
                return Type('char8', **params)    # IEC 9899 defines sizeof(char)==1
            if _float_match(s):
                return Type('float32', **params)  # IEC 60559 defines sizeof(float)==4
            if _double_match(s):
                return Type('float64', **params)  # IEC 60559 defines sizeof(double)==8
            target_info = TargetInfo()
            for match, otype, ntype in [
                    (_wchar_match, 'wchar', 'char'),

                    (_schar_match, 'char', 'int'),
                    (_uchar_match, 'uchar', 'uint'),

                    (_byte_match, 'byte', 'int'),
                    (_ubyte_match, 'ubyte', 'uint'),

                    (_short_match, 'short', 'int'),
                    (_ushort_match, 'ushort', 'uint'),

                    (_int_match, 'int', 'int'),
                    (_uint_match, 'uint', 'uint'),

                    (_long_match, 'long', 'int'),
                    (_ulong_match, 'ulong', 'uint'),

                    (_longlong_match, 'longlong', 'int'),
                    (_ulonglong_match, 'ulonglong', 'uint'),

                    (_size_t_match, 'size_t', 'uint'),
                    (_ssize_t_match, 'ssize_t', 'int'),

                    (_longdouble_match, 'longdouble', 'float'),
                    (_complex_match, 'complex', 'complex'),
            ]:
                if match(s) is not None:
                    sz = target_info.sizeof(otype)
                    if sz is not None:
                        bits = str(sz * 8)
                        return Type(ntype + bits, **params)
                    break
            if target_info.strict:
                raise ValueError('%s is not concrete' % (self))
            return self
        if self.is_pointer:
            return Type(
                self[0]._normalize(), self[1], **params)
        if self.is_struct:
            return Type(
                *(t._normalize() for t in self), **params)
        if self.is_function:
            return Type(
                self[0]._normalize(),
                tuple(t._normalize() for t in self[1]),
                **params)
        if self.is_custom:
            inner = []
            for a in self[0]:
                if isinstance(a, Type):
                    a = a._normalize()
                inner.append(a)
            return type(self)(tuple(inner), **params)
        raise NotImplementedError(repr(self))

    def mangle(self):
        """Return mangled type string.

        Mangled type string is a string representation of the type
        that can be used for extending the function name.
        """
        name_suffix = ''
        name = self.name
        if name:
            name_suffix = 'W' + str(len(name)) + 'W' + name
        if self.is_void:
            assert name_suffix == '', name_suffix
            return 'v'
        if self.is_pointer:
            return '_' + self[0].mangle() + 'P' + name_suffix
        if self.is_struct:
            return '_' + ''.join(m.mangle()
                                 for m in self) + 'K' + name_suffix
        if self.is_function:
            r = self[0].mangle()
            a = ''.join([a.mangle() for a in self[1]])
            return '_' + r + 'a' + a + 'A' + name_suffix
        if self.is_custom:
            params = self[0]
            if type(self) is Type:
                name = params[0]
                params = params[1:]
            else:
                name = type(self).__name__
            n = _mangling_map.get(name)
            n = name if n is None else n
            r = 'V' + str(len(n)) + 'V' + n
            a = ''.join([a.mangle() for a in params])
            return '_' + r + 'r' + a + 'R' + name_suffix
        if self.is_atomic:
            n = _mangling_map.get(self[0])
            if n is not None:
                return n + name_suffix
            n = self[0]
            return 'V' + str(len(n)) + 'V' + n + name_suffix
        raise NotImplementedError(repr(self))

    @classmethod
    def demangle(cls, s):
        block, rest = _demangle(s)
        assert not rest, repr(rest)
        assert len(block) == 1, repr(block)
        return block[0]

    @property
    def bits(self):
        if self.is_void:
            return 0
        if self.is_bool:
            return int(self[0][4:] or 1)
        if self.is_int:
            return int(self[0][3:])
        if self.is_uint or self.is_char:
            return int(self[0][4:])
        if self.is_float:
            return int(self[0][5:])
        if self.is_complex:
            return int(self[0][7:])
        if self.is_struct:
            return sum([m.bits for m in self])
        if self.is_pointer:
            return TargetInfo().sizeof('voidptr')*8 or 64
        return NotImplemented

    def match(self, other):
        """Return match penalty when other can be converted to self.
        Otherwise, return None.

        Parameters
        ----------
        other : {Type, tuple}
          Specify other signature. If other is a tuple of signatures,
          then it is interpreted as argument types of a function
          signature.

        Returns
        -------
        penalty : {int, None}
          Penalty of a match. For a perfect match, penalty is 0.
          If match is impossible, return None
        """
        if isinstance(other, Type):
            if self == other:
                return 0
            if other.is_void:
                return (0 if self.is_void else None)
            elif self.is_custom:
                if type(self) is not type(other):
                    return
                if self[0] == other[0]:
                    if self._params or other._params:
                        warnings.warn(
                            'The default match implementation ignores _params content. Please'
                            f' implement match method for custom type {type(self).__name__}')
                    return 1
                return
            elif other.is_pointer:
                if not self.is_pointer:
                    return
                if self[0].is_void or other[0].is_void:
                    return 0
                penalty = self[0].match(other)
                if penalty is None:
                    if self[0].is_void:
                        penalty = 1
                return penalty
            elif self.is_pointer:
                return
            elif other.is_struct:
                if not self.is_struct:
                    return
                if len(self) != len(other):
                    return
                penalty = 0
                for a, b in zip(self, other):
                    p = a.match(b)
                    if p is None:
                        return
                    penalty = penalty + p
                return penalty
            elif self.is_struct:
                return
            elif other.is_function:
                if not self.is_function:
                    return
                if len(self[1]) != len(other[1]):
                    return
                penalty = self[0].match(other[0])
                if penalty is None:
                    return
                for a, b in zip(self[1], other[1]):
                    p = a.match(b)
                    if p is None:
                        return
                    penalty = penalty + p
                return penalty
            elif self.is_function:
                return
            elif (
                    (other.is_int and self.is_int)
                    or (other.is_bool and self.is_bool)
                    or (other.is_float and self.is_float)
                    or (other.is_uint and self.is_uint)
                    or (other.is_char and self.is_char)
                    or (other.is_complex and self.is_complex)):
                if self.bits >= other.bits:
                    return 0
                return other.bits - self.bits
            elif (
                    (other.is_int and self.is_uint)
                    or (other.is_uint and self.is_int)):
                if self.bits >= other.bits:
                    return 10
                return 10 + other.bits - self.bits
            elif self.is_complex and (other.is_float
                                      or other.is_int or other.is_uint):
                return 1000
            elif self.is_float and (other.is_int or other.is_uint):
                return 1000
            elif (self.is_float or self.is_complex) and other.is_bool:
                return
            elif (self.is_int or self.is_uint) and other.is_bool:
                if self.bits >= other.bits:
                    return 0
                return other.bits - self.bits
            elif self.is_bool and (other.is_int or other.is_uint):
                if self.bits >= other.bits:
                    return 0
                return other.bits - self.bits
            elif ((self.is_int or self.is_uint or self.is_bool)
                  and (other.is_float or other.is_complex)):
                return
            elif self.is_float and other.is_complex:
                return
            elif self.is_pointer and (other.is_int or other.is_uint):
                if self.bits == other.bits:
                    return 1
                return
            elif ((self.is_string and not other.is_string)
                  or (not self.is_string and other.is_string)):
                return

            # TODO: lots of
            raise NotImplementedError(repr((self, other)))
        elif isinstance(other, tuple):
            if not self.is_function:
                return
            atypes = self[1]
            if len(other) == 0 and len(atypes) == 1 and atypes[0].is_void:
                return 0
            if len(atypes) != len(other):
                return
            penalty = 0
            for a, b in zip(atypes, other):
                p = a.match(b)
                if p is None:
                    return
                penalty = penalty + p
            return penalty
        raise NotImplementedError(repr(type(other)))

    def __call__(self, *atypes, **params):
        return Type(self, atypes or (Type(),), **params)

    def pointer(self):
        numba_ptr_type = self._params.get('NumbaPointerType')
        if numba_ptr_type is not None:
            return Type(self, '*').params(NumbaType=numba_ptr_type)
        return Type(self, '*')

    def apply_templates(self, templates):
        """Iterator of concrete types derived from applying templates mapping
        to self.

        The caller must allow templates dictionary to be changed in-situ.
        """
        if self.is_concrete:
            yield self
        elif self.is_atomic:
            type_list = templates.get(self[0])
            if type_list:
                assert isinstance(type_list, list), type_list
                for i, t in enumerate(type_list):
                    # using copy so that template symbols will not be
                    # contaminated with self parameters
                    t = Type.fromobject(t)
                    t_ = t.copy()
                    if t.is_concrete:
                        # templates is changed in-situ! This ensures that
                        # `T(T)` produces `i4(i4)`, `i8(i8)` for `T in
                        # [i4, i8]`. To produce all possible combinations
                        # `i4(i4)`, `i8(i8)`, `i8(i4)`, `i4(i8)`, use
                        # `T1(T2)` where `T1 in [i4, i8]` and `T2 in [i4,
                        # i8]`
                        templates[self[0]] = [t]
                        yield t_.params(None, **self._params)
                        # restore templates
                        templates[self[0]] = type_list
                    else:
                        # this will avoid infinite recursion when
                        # templates is `T in [U]` and `U in [T]`
                        templates[self[0]] = []
                        for ct in t.apply_templates(templates):
                            templates[self[0]] = [ct]
                            yield ct.params(None, **self._params)
                            templates[self[0]] = []
                        templates[self[0]] = type_list
            else:
                raise TypeError(f'cannot make {self} concrete using template mapping {templates}')
        elif self.is_pointer:
            for t in self[0].apply_templates(templates):
                yield Type(t, self[1], **self._params)
        elif self.is_struct:
            for i, t in enumerate(self):
                if not t.is_concrete:
                    for ct in t.apply_templates(templates):
                        yield from Type(
                            *(self[:i] + (ct,) + self[i+1:]),
                            **self._params).apply_templates(templates)
                    return
            assert 0
        elif self.is_function:
            rtype, atypes = self
            if not rtype.is_concrete:
                for rt in rtype.apply_templates(templates):
                    yield from Type(rt, atypes, **self._params).apply_templates(templates)
                return
            for i, t in enumerate(atypes):
                if not t.is_concrete:
                    for ct in t.apply_templates(templates):
                        yield from Type(
                            rtype, atypes[:i] + (ct,) + atypes[i+1:],
                            **self._params).apply_templates(templates)
                    return
            assert 0
        elif self.is_custom:
            params = self[0]
            cls = type(self)
            if cls is Type:
                name = params[0]
                params = params[1:]
            else:
                name = cls.__name__
            if name in templates:
                for cname in templates[name]:
                    cname = Type.aliases.get(cname, cname)
                    if isinstance(cname, str):
                        cls = Type.custom_types.get(cname, Type)
                    else:
                        assert issubclass(cname, Type), cname
                        cls = cname
                        cname = cls.__name__
                    if cls is Type:
                        typ = cls((cname,) + params)
                    else:
                        typ = cls(params)
                    yield from typ.params(self).apply_templates(templates)
                return
            for i, t in enumerate(params):
                if not isinstance(t, Type):
                    continue
                if not t.is_concrete:
                    for ct in t.apply_templates(templates):
                        if cls is Type:
                            yield from cls(
                                (name,) + params[:i] + (ct,) + params[i+1:],
                                **self._params).apply_templates(templates)
                        else:
                            yield from cls(
                                params[:i] + (ct,) + params[i+1:],
                                **self._params).apply_templates(templates)
                    return
            assert 0, repr(self)
        else:
            raise NotImplementedError(f'apply_templates: {self} {templates}')


def _demangle_name(s, suffix='W'):
    if s and s[0] == suffix:
        i = s.find(suffix, 1)
        assert i != -1, repr(s)
        ln = int(s[1:i])
        rest = s[i+ln+1:]
        return s[i+1:i+ln+1], rest
    return None, s


def _demangle(s):
    """Helper function to demangle the string of mangled Type.

    Used internally.

    Algorithm invented by Pearu Peterson, February 2019
    """
    if not s:
        return (Type(),), ''
    if s[0] == 'V':
        name, rest = _demangle_name(s, suffix='V')
        typ = Type(name)
    elif s[0] == '_':
        block, rest = _demangle(s[1:])
        kind, rest = rest[0], rest[1:]
        assert kind in '_'+_mangling_suffices+_mangling_prefixes, repr(kind)
        if kind == 'P':
            assert len(block) == 1, repr(block)
            typ = Type(block[0], '*')
        elif kind == 'K':
            typ = Type(*block)
        elif kind == 'a':  # function
            assert len(block) == 1, repr(block)
            rtype = block[0]
            atypes, rest = _demangle('_' + rest)
            typ = Type(rtype, atypes, name='')
        elif kind in 'AR':
            return block, rest
        elif kind == 'r':  # custom
            assert len(block) == 1, repr(block)
            name = block[0]
            assert name.is_atomic, name
            name = name[0]
            if rest and rest[0] == 'R':
                atypes = ()
                rest = rest[1:]
            else:
                atypes, rest = _demangle('_' + rest)
            if name in Type.custom_types:
                typ = Type.custom_types[name](atypes)
            else:
                typ = Type((name,)+atypes)
        else:
            raise NotImplementedError(repr((kind, s)))
    else:
        rest = s[1:]
        t = _mangling_imap[s[0]]
        if t == 'void':
            typ = Type()
        else:
            typ = Type(t)
    vname, rest = _demangle_name(rest)
    if vname is not None:
        typ._params['name'] = vname
    result = [typ]
    if rest and rest[0] not in _mangling_prefixes:
        r, rest = _demangle(rest)
        result.extend(r)
    return tuple(result), rest


# TODO: move numba boolean support to rbc/boolean_type.py

class Boolean1(nb.types.Boolean):

    def can_convert_from(self, typingctx, other):
        return isinstance(other, nb.types.Boolean)


@datamodel.register_default(Boolean1)
class Boolean1Model(datamodel.models.BooleanModel):

    def get_data_type(self):
        return self._bit_type


class Boolean8(nb.types.Boolean):

    bitwidth = 8

    def can_convert_to(self, typingctx, other):
        return isinstance(other, nb.types.Boolean)

    def can_convert_from(self, typingctx, other):
        return isinstance(other, nb.types.Boolean)


@datamodel.register_default(Boolean8)
class Boolean8Model(datamodel.models.BooleanModel):

    def get_value_type(self):
        return self._byte_type


boolean1 = Boolean1('boolean1')
boolean8 = Boolean8('boolean8')


@lower_cast(Boolean1, nb.types.Boolean)
@lower_cast(Boolean8, nb.types.Boolean)
def literal_booleanN_to_boolean(context, builder, fromty, toty, val):
    return builder.icmp_signed('!=', val, val.type(0))


@lower_cast(nb.types.Boolean, Boolean1)
@lower_cast(nb.types.Boolean, Boolean8)
def literal_boolean_to_booleanN(context, builder, fromty, toty, val):
    llty = context.get_value_type(toty)
    return builder.zext(val, llty)


@extending.lower_builtin(bool, Boolean8)
def boolean8_to_bool(context, builder, sig, args):
    [val] = args
    return builder.icmp_signed('!=', val, val.type(0))


_numba_bool_map[1] = boolean1
_numba_bool_map[8] = boolean8
_numba_imap[boolean1] = 'bool1'
_numba_imap[boolean8] = 'bool8'

boolean8ptr = nb.types.CPointer(boolean8)
boolean8ptr2 = nb.types.CPointer(boolean8ptr)

_pointer_types = [boolean8ptr, boolean8ptr2, nb.types.intp, nb.types.voidptr]
for _i, _p1 in enumerate(_pointer_types[:2]):
    for _j, _p2 in enumerate(_pointer_types):
        if _p1 == _p2 or _i > _j:
            continue
        typeconv.rules.default_type_manager.set_compatible(
            _p1, _p2, typeconv.Conversion.safe)
        typeconv.rules.default_type_manager.set_compatible(
            _p2, _p1, typeconv.Conversion.safe)


_ufunc_pos_args_match = re.compile(
    r'(?P<name>\w[\w\d_]*)\s*[(](?P<pos_args>[^/)]*)[/]?(?P<rest>.*)[)]').match
_req_opt_args_match = re.compile(r'(?P<req_args>[^[]*)(?P<opt_args>.*)').match
_annot_match = re.compile(
    r'\s*(?P<name>\w[\w\d_]*)\s*[:](?P<annotation>.*)').match


def get_signature(obj):
    if inspect.isfunction(obj):
        return inspect.signature(obj)
    if isinstance(obj, np.ufunc):
        parameters = dict()
        returns = dict()

        sigline = obj.__doc__.lstrip().splitlines(1)[0]
        m = _ufunc_pos_args_match(sigline)
        name = m['name']
        assert name == obj.__name__, (name, obj.__name__)

        # positional arguments
        m = _req_opt_args_match(m['pos_args'])
        req_pos_names, opt_pos_names = [], []
        for n in m.group('req_args').split(','):
            n = n.strip()
            if n:
                req_pos_names.append(n)
                parameters[n] = inspect.Parameter(
                    n, inspect.Parameter.POSITIONAL_ONLY)
        for n in (m.group('opt_args').replace('[', '')
                   .replace(']', '').split(',')):
            n = n.strip()
            if n:
                opt_pos_names.append(n)
                parameters[n] = inspect.Parameter(
                    n, inspect.Parameter.POSITIONAL_ONLY, default=None)
        # TODO: process non-positional arguments in `m['rest']`

        # scan for annotations and determine returns
        mode = 'none'
        for line in obj.__doc__.splitlines():
            if line in ['Parameters', 'Returns', 'Notes', 'See Also',
                        'Examples']:
                mode = line
                continue
            if mode == 'Parameters':
                m = _annot_match(line)
                if m is not None:
                    n = m['name']
                    if n in parameters:
                        annot = m['annotation'].strip()
                        parameters[n] = parameters[n].replace(
                            annotation=annot)
            if mode == 'Returns':
                m = _annot_match(line)
                if m is not None:
                    n = m['name']
                    annot = m['annotation'].strip()
                    returns[n] = annot

        sig = inspect.Signature(parameters=parameters.values())
        if len(returns) == 1:
            sig = sig.replace(return_annotation=list(returns.values())[0])
        elif returns:
            sig = sig.replace(return_annotation=returns)
        return sig

    raise NotImplementedError(obj)


# Import numba support

if 1:  # to avoid flake E402
    from . import structure_type



__all__ = ['OmnisciTableFunctionManagerType']


from numba.core import extending, types
from numba.core.cgutils import make_bytearray, global_constant
from rbc import structure_type, irutils
from rbc.errors import UnsupportedError
from rbc.targetinfo import TargetInfo
from rbc.typesystem import Type
from llvmlite import ir


class OmnisciTableFunctionManagerType(Type):
    """TableFunctionManager<> is a typesystem custom type that
    represents a class type with the following public interface:

      class TableFunctionManager
      {
        int32_t error_message(const char* message);
        void set_output_row_size(int64_t num_rows);
      }

    """

    @property
    def __typesystem_type__(self):
        ptr_t = Type.fromstring("int8 ptr")
        return Type(ptr_t).params(NumbaPointerType=OmnisciTableFunctionManagerNumbaType).pointer()


class OmnisciTableFunctionManagerNumbaType(structure_type.StructureNumbaPointerType):
    pass


error_msg = 'TableFunctionManager is only available in OmniSciDB 5.9 or newer (got %s)'
i8p = ir.IntType(8).as_pointer()
i32 = ir.IntType(32)
i64 = ir.IntType(64)


@extending.intrinsic
def omnisci_udtfmanager_error_message_(typingctx, mgr, msg):
    sig = types.int32(mgr, msg)

    target_info = TargetInfo()
    if target_info.software[1][:3] < (5, 9, 0):
        raise UnsupportedError(error_msg % (".".join(map(str, target_info.software[1]))))

    if not isinstance(msg, types.StringLiteral):
        raise TypeError(f"expected StringLiteral but got {type(msg).__name__}")

    def codegen(context, builder, signature, args):
        mgr_ptr = args[0]

        mgr_i8ptr = builder.bitcast(mgr_ptr, i8p)

        msg_bytes = msg.literal_value.encode('utf-8')
        msg_const = make_bytearray(msg_bytes + b'\0')
        msg_global_var = global_constant(
            builder.module,
            "table_function_manager_error_message",
            msg_const)
        msg_ptr = builder.bitcast(msg_global_var, i8p)

        fnty = ir.FunctionType(i32, [i8p, i8p])
        fn = irutils.get_or_insert_function(
            builder.module, fnty, "TableFunctionManager_error_message")

        return builder.call(fn, [mgr_i8ptr, msg_ptr])

    return sig, codegen


@extending.overload_method(OmnisciTableFunctionManagerNumbaType, 'error_message')
def omnisci_udtfmanager_error_message(mgr, msg):
    def impl(mgr, msg):
        return omnisci_udtfmanager_error_message_(mgr, msg)
    return impl


@extending.intrinsic
def omnisci_udtfmanager_set_output_row_size_(typingctx, mgr, num_rows):
    sig = types.void(mgr, num_rows)

    target_info = TargetInfo()
    if target_info.software[1][:3] < (5, 9, 0):
        raise UnsupportedError(error_msg % (".".join(map(str, target_info.software[1]))))

    def codegen(context, builder, sig, args):
        mgr_ptr, num_rows_arg = args

        mgr_i8ptr = builder.bitcast(mgr_ptr, i8p)

        fnty = ir.FunctionType(ir.VoidType(), [i8p, i64])
        fn = irutils.get_or_insert_function(
            builder.module, fnty, "TableFunctionManager_set_output_row_size")

        builder.call(fn, [mgr_i8ptr, num_rows_arg])

    return sig, codegen


@extending.overload_method(OmnisciTableFunctionManagerNumbaType, 'set_output_row_size')
def omnisci_udtfmanager_set_output_row_size(mgr, num_rows):
    def impl(mgr, num_rows):
        return omnisci_udtfmanager_set_output_row_size_(mgr, num_rows)
    return impl



from rbc.targetinfo import TargetInfo


def test_basic():
    host = TargetInfo.host()
    assert host.name == 'host_cpu'
    with host:
        ti = TargetInfo()
        assert ti is host


def test_to_from_dict():
    host = TargetInfo.host()
    d = host.todict()
    host2 = TargetInfo.fromdict(d)
    assert host.__dict__ == host2.__dict__


def test_host_cache():
    host = TargetInfo.host()
    host2 = TargetInfo.host()
    assert host is host2


def test_allocation_functions():
    host = TargetInfo.host()
    dhost = TargetInfo.host(use_tracing_allocator=True)
    assert host is not dhost  # check that use_tracing_allocator is part of the cache key
    assert host.info['fn_allocate_varlen_buffer'] == 'rbclib_allocate_varlen_buffer'
    assert host.info['fn_free_buffer'] == 'rbclib_free_buffer'
    #
    assert dhost.info['fn_allocate_varlen_buffer'] == 'rbclib_tracing_allocate_varlen_buffer'
    assert dhost.info['fn_free_buffer'] == 'rbclib_tracing_free_buffer'



"""
rbclib is a C runtime library for rbc.

Sometimes the IR code generated by rbc needs to call some external helper
function written in C; rbclib makes it possible to write such functions.

This is the file layout:

  - _rbclib.c contains the actual C source code

  - _rbclib_build.py is the CFFI build script which is called by setup.py

  - _rbclib.abi3.so (_rbclib.pyd on Windows) is the native extension produced
    by _rbclib_build.py and that can be imported by 'import rbc.rbclib._rbclib'

rbclib functions can be called in two ways:

  1. From pure Python, through CFFI: for this you need to call
     e.g. lib._rbclib_add_ints(). This is useful e.g. inside tests

  2. From an @rjit function, through rbc.  For this you need to call
     e.g. rbclib.add_ints. The codegen logic is in rbclib/intrinsic.py.  In
     this case, the generated LLVM will contain a direct call to the
     underlying C function (_rbclib_add_ints), which means that such a
     function must be loaded in memory and LLVM must be aware of it. This is
     done by load_inside_llvm(), which is called from
     libfuncs.RBCLib.__init__().

NOTE: rbclib is based on CFFI, but this is not strictly necessary: in theory,
it is enough to dlopen() a C library which contains the desired
functions. However, using CFFI has many advantages:

  1. It is very easy to integrate the build process with setup.py

  2. You can call the functions also from Python (useful for testing)

  3. Loading the shared library in memory is as simple as importing the
     _rbclib module, instead of having to load the library explicitly
     e.g. through ctypes
"""

import llvmlite.binding
try:
    from . import _rbclib
except ImportError as e:
    # improve the ImportError error message
    msg = (f"{e}\nThis probably indicates "
           "that rbc has not been built/installed correctly, possibly "
           "because cffi was not available at compilation time")
    e.msg = msg
    e.args = (msg,)
    raise

from ._rbclib import lib, ffi  # noqa: F401
from .intrinsic import add_ints  # noqa: F401
from . import tracing_allocator  # noqa: F401, side effects

# this contains all the C names which we want to expose to the RBC
# compiler. See libfuncs.RBCLib.
FUNCTION_NAMES = [
    '_rbclib_add_ints',
    'rbclib_allocate_varlen_buffer',
    'rbclib_free_buffer',
    'rbclib_tracing_allocate_varlen_buffer',
    'rbclib_tracing_free_buffer',
]


def load_inside_llvm():
    llvmlite.binding.load_library_permanently(_rbclib.__file__)



import pytest
from functools import reduce
from typing import Tuple
from rbc.tests import omnisci_fixture
from rbc.externals import libdevice

libdevicefuncs = pytest.importorskip("numba.cuda.libdevicefuncs")

funcs = []
for fname, (retty, args) in libdevicefuncs.functions.items():
    assert fname.startswith('__nv_'), fname
    argtys = tuple(map(lambda x: str(x.ty), args))
    arity = len(argtys)
    has_ptr_arg = reduce(lambda x, y: x or y.is_ptr, args, False)
    funcs.append((fname, str(retty), argtys, has_ptr_arg))


@pytest.fixture(scope="module")
def omnisci():

    for o in omnisci_fixture(globals()):
        if not o.has_cuda:
            pytest.skip("cuda is not enabled")
        define(o)
        yield o


def define(omnisci):
    def inner(fname: str, retty: str, argtypes: Tuple[str]):
        cmath_fn = getattr(libdevice, fname)
        arity = len(argtypes)

        # define omnisci callable
        if arity == 1:

            def fn(a):
                return cmath_fn(a)

        elif arity == 2:

            def fn(a, b):
                return cmath_fn(a, b)

        else:

            def fn(a, b, c):
                return cmath_fn(a, b, c)

        fn.__name__ = f"{omnisci.table_name}_{fname[5:]}"
        fn = omnisci(f"{retty}({', '.join(argtypes)})", devices=["gpu"])(fn)

    for fname, retty, argtys, has_ptr_arg in funcs:
        if has_ptr_arg:
            continue
        inner(fname, str(retty), argtys)


cols_dict = {
    "float32": "f4",
    "float64": "f8",
    "int16": "i2",
    "int32": "i4",
    "int64": "i8",
}


@pytest.mark.slow
@pytest.mark.parametrize(
    "fname,retty,argtys,has_ptr_arg", funcs, ids=[item[0] for item in funcs]
)
def test_externals_libdevice(omnisci, fname, retty, argtys, has_ptr_arg):
    if has_ptr_arg:
        pytest.skip(f"{fname} has a pointer argument")

    func_name = f"{omnisci.table_name}_{fname[5:]}"
    table = f"{omnisci.table_name}"

    if fname[5:] in ["brev", "brevll"]:
        query = f"SELECT {func_name}(i4+4) FROM {table}"
    elif fname[5:] in ["ilogb", "ilogbf"]:
        query = f"SELECT {func_name}({cols_dict[argtys[0]]}+2.0) FROM {table}"
    else:
        cols = ", ".join(tuple(map(lambda x: cols_dict[x], argtys)))
        query = f"SELECT {func_name}({cols}) FROM {table}"

    _, _ = omnisci.sql_execute(query)



"""Client implements API for calling methods defined in multiplex
thrift server. The clients thrift configuration will be downloaded
from the server (the server has to implement a service `info` with a
method `thrift_content`).

"""
# Author: Pearu Peterson
# Created: February 2019

import os
import tempfile
import warnings
from threading import Lock
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import thriftpy2 as thr
    import thriftpy2.rpc
import pickle
import six
from . import types


# lock to prevent more than one thread from initializing the thrift client
lock = Lock()


def from_thrift(thrift, spec, result):
    if not spec.thrift_spec:
        assert result is None, repr(type(result))
        return result
    if 0 not in spec.thrift_spec:
        return result
    t = spec.thrift_spec[0]
    if t[0] == thr.thrift.TType.STRING:
        assert isinstance(result, str), repr(type(result))
        return result
    elif t[0] in [
            thr.thrift.TType.I08,
            thr.thrift.TType.I16,
            thr.thrift.TType.I32,
            thr.thrift.TType.I64]:
        assert isinstance(result, int), repr(type(result))
        return result
    elif t[0] == thr.thrift.TType.BOOL:
        assert isinstance(result, bool), repr(type(result))
        return result
    elif t[0] == thr.thrift.TType.DOUBLE:
        assert isinstance(result, float), repr(type(result))
        return result
    elif t[0] == thr.thrift.TType.SET:
        return set(result)
    elif t[0] == thr.thrift.TType.LIST:
        assert isinstance(result, list), repr(type(result))
        return result
    elif t[0] == thr.thrift.TType.MAP:
        assert isinstance(result, dict), repr(type(result))
        return result
    elif t[0] == thr.thrift.TType.STRUCT:
        # _i, status, tcls, _b = t
        # assert t[1] == 'success', repr(t)
        # assert isinstance(result, t[2]), repr(type(result))
        return types.toobject(thrift, result)

    raise NotImplementedError(repr((t, type(result))))


class Client(object):
    """Thrift multiplex client.

    The thrift client loads the thrift configuration from a thrift
    server.
    """

    def __init__(self, thrift_content=None, **options):
        self.thrift_content = thrift_content
        self.multiplexed = options.pop('multiplexed', True)
        self.thrift_content_service = options.pop(
            'thrift_content_service', 'info')
        self.options = options
        with lock:
            self._update_thrift()

    def _update_thrift(self):
        """Update client thrift configuration from server.

        The servers thrift configuration must contain at least

          service info { string thrift_content() }

        and the server thrift must run in multiplexed mode.
        """
        if self.thrift_content is None:
            i, fn = tempfile.mkstemp(suffix='.thrift', prefix='rpc-client0-')
            f = os.fdopen(i, mode='w')
            f.write('service %s { string thrift_content() }'
                    % (self.thrift_content_service))
            f.close()
            tmp_thrift = thr.load(fn)
            os.remove(fn)
            if not self.multiplexed:
                factory = thr.protocol.TBinaryProtocolFactory()
            else:
                factory = thr.protocol.TMultiplexedProtocolFactory(
                    thr.protocol.TBinaryProtocolFactory(),
                    self.thrift_content_service)
            service = getattr(tmp_thrift, self.thrift_content_service)
            ctx = thr.rpc.client_context(service,
                                         proto_factory=factory,
                                         **self.options)
            with ctx as c:
                self.thrift_content = c.thrift_content()
        i, fn = tempfile.mkstemp(suffix='.thrift', prefix='rpc-client-')
        f = os.fdopen(i, mode='w')
        f.write(self.thrift_content)
        f.close()
        self.thrift = thr.load(fn)
        os.remove(fn)

    def _args_to_thrift(self, spec, args):
        thrift_spec = spec.thrift_spec
        thrift_args = []
        for i, arg in enumerate(args):
            t = thrift_spec[i+1]
            if t[0] == thr.thrift.TType.STRING:
                arg = types.fromobject(self.thrift, str, arg)
            elif t[0] == thr.thrift.TType.STRUCT:
                arg = types.fromobject(self.thrift, t[2], arg)
            elif t[0] in [
                    thr.thrift.TType.I08,
                    thr.thrift.TType.I16,
                    thr.thrift.TType.I32,
                    thr.thrift.TType.I64]:
                arg = types.fromobject(self.thrift, int, arg)
            elif t[0] == thr.thrift.TType.BOOL:
                arg = types.fromobject(self.thrift, bool, arg)
            elif t[0] == thr.thrift.TType.DOUBLE:
                arg = types.fromobject(self.thrift, float, arg)
            elif t[0] == thr.thrift.TType.SET:
                arg = types.fromobject(self.thrift, set, arg)
            elif t[0] == thr.thrift.TType.LIST:
                arg = types.fromobject(self.thrift, list, arg)
            elif t[0] == thr.thrift.TType.MAP:
                arg = types.fromobject(self.thrift, dict, arg)
            else:
                raise NotImplementedError(repr((t, type(arg))))
            thrift_args.append(arg)
        return tuple(thrift_args)

    def _result_from_thrift(self, spec, result):
        return from_thrift(self.thrift, spec, result)

    def __call__(self, **services):
        """Perform a RPC call to thrift server.

        Parameters
        ----------
        services : dict
          Specify a service mapping with arguments to server methods.

        Returns
        -------
        results : dict
          The return values of server methods organized in a service
          mapping. See example below.

        Example
        -------

        1. Create a connection to thrift server:

          conn = Client(host=.., port=..)

        2. Say, a server implements two services, Ping and King. Ping
        service has a method `bong(i) -> int` and King has a method
        `kong(s) -> str`. To call these methods from a client, use

          results = conn(Ping=dict(bong=(5,)), King=dict(kong=('hello')))

        The `results` will be, for instance,

          dict(Ping=dict(bong=6), King=dict(kong='hi'))

        where `bong(5) -> 6` and `kong('hello') -> 'hi'` is assumed.
        """
        exception = getattr(self.thrift, 'Exception', Exception)

        results = {}
        for service_name, query_dict in services.items():
            factory = thr.protocol.TBinaryProtocolFactory()
            if self.multiplexed:
                factory = thr.protocol.TMultiplexedProtocolFactory(
                    factory, service_name)
            service = getattr(self.thrift, service_name)
            ctx = thr.rpc.client_context(service, proto_factory=factory,
                                         **self.options)
            results[service_name] = {}
            with ctx as c:
                for query_name, query_args in query_dict.items():
                    argsmth = getattr(service, query_name + '_args')
                    query_args = self._args_to_thrift(argsmth, query_args)
                    mth = getattr(c, query_name)
                    assert mth is not None
                    exc = None
                    try:
                        r = mth(*query_args)
                    except exception as msg:
                        if exception is Exception:
                            raise
                        if msg.kind == self.thrift.ExceptionKind.EXC_TBLIB:
                            exc = pickle.loads(msg.message)
                        elif msg.kind == self.thrift.ExceptionKind.EXC_MESSAGE:
                            et, ev = msg.message.split(':', 1)
                            et = __builtins__.get(et)
                            if et is None:
                                et = Exception
                                ev = msg.message
                            exc = et, et(ev.lstrip()), None
                        else:
                            raise
                    except Exception as msg:
                        print(msg)
                        raise
                    if exc is not None:
                        six.reraise(*exc)   # RAISING SERVER EXCEPTION
                    r = self._result_from_thrift(
                        getattr(service, query_name + '_result'), r)
                    results[service_name][query_name] = r
        return results



__all__ = ['omnisci_fixture', 'sql_execute']


import os
import pytest
import warnings
import numpy
from collections import defaultdict


def assert_equal(actual, desired):
    """Test equality of actual and desired.

    When both inputs are numpy array or number objects, test equality
    of dtype attributes as well.
    """
    numpy.testing.assert_equal(actual, desired)

    if isinstance(actual, numpy.ndarray) and isinstance(desired, numpy.ndarray):
        numpy.testing.assert_equal(actual.dtype, desired.dtype)
    elif isinstance(actual, numpy.number) and isinstance(desired, numpy.number):
        numpy.testing.assert_equal(actual.dtype, desired.dtype)


def sql_execute(query):
    """Execute a SQL statement to omniscidb server using global instance.

    Use when the query does not require registration of new UDF/UDTFs.
    """
    rbc_omnisci = pytest.importorskip('rbc.omniscidb')
    omnisci = next(rbc_omnisci.global_omnisci_singleton)
    return omnisci.sql_execute(query)


def omnisci_fixture(caller_globals, minimal_version=(0, 0),
                    suffices=['', '10', 'null', 'array', 'arraynull'],
                    load_columnar=True, load_test_data=True, debug=False):
    """Usage from a rbc/tests/test_xyz.py file:

    .. code-block:: python

       import pytest
       from rbc.tests import omnisci_fixture

       @pytest.fixture(scope='module')
       def omnisci():
           from o in omnisci_fixture(globals()):
               # do some customization here
               yield o

    This fixture creates the following tables:

    f'{omnisci.table_name}' - contains columns f8, f4, i8, i4, i2, i1,
                              b with row size 5.

    f'{omnisci.table_name}10' - contains columns f8, f4, i8, i4, i2,
                                i1, b with row size 10.

    f'{omnisci.table_name}null' - contains columns f8, f4, i8, i4, i2,
                                  i1, b with row size 5, contains null
                                  values

    f'{omnisci.table_name}array' - contains arrays f8, f4, i8, i4, i2,
                                   i1, b with row size 5

    f'{omnisci.table_name}arraynull' - contains arrays f8, f4, i8, i4, i2,
                                       i1, b with row size 5, contains null
                                       values.
    """
    rbc_omnisci = pytest.importorskip('rbc.omniscidb')
    available_version, reason = rbc_omnisci.is_available()

    def require_version(version, message=None, label=None):
        """Execute pytest.skip(...) if version is older than available_version.

        Some tests can be run only when using omniscidb server built
        from a particular branch of omniscidb.  So, when the specified
        version and the omniscidb version match exactly and these
        correspond to the current development version, if the
        specified label does not match with the value of envrinment
        variable OMNISCIDB_DEV_LABEL, then the corresponing test will
        be skipped. Use label 'docker-dev' when using omniscidb dev
        docker image.

        """
        # The available version (of the omniscidb server) has date and
        # hash bits, however, these are useless for determining the
        # version ordering (in the case of available_version[:3] ==
        # version) because the date corresponds to the date of
        # building the server and the hash corresponds to some commit
        # of some repository (omniscidb or omniscidb-internal) and it
        # does not provide easy date information.
        #
        # The condition available_version[:3] == version can appear in
        # the following cases (given in the order of from newer to
        # older):
        # 1. omniscidb is built against a omniscidb-internal PR branch
        # (assuming it is rebased against master)
        # 2. omniscidb is built against omniscidb-internal master branch
        # 3. omniscidb is built against omniscidb master branch
        # 4. omniscidb is built against omniscidb dev docker
        # 5. omniscidb is built against omniscidb/omniscidb-internal release tag
        #
        # rbc testing suite may use features that exists in the head
        # of the above list but not in the tail of it. So we have a
        # problem of deciding if a particular test should be disabled
        # or not for a given case while there is no reliable way to
        # tell from omniscidb version if the server has the particular
        # feature or not. To resolve this, we use label concept as
        # explained in the doc-string.
        #

        if not available_version:
            pytest.skip(reason)

        # Requires update when omniscidb-internal bumps up version number:
        current_development_version = (6, 0, 0)
        if available_version[:3] > current_development_version:
            warnings.warn(f'{available_version}) is newer than development version'
                          f' ({current_development_version}), please update the latter!')

        assert isinstance(version, tuple)
        if version > available_version[:3]:
            _reason = f'test requires version {version} or newer, got {available_version}'
            if message is not None:
                _reason += f': {message}'
            pytest.skip(_reason)

        if label is not None:
            env_label = os.environ.get('OMNISCIDB_DEV_LABEL')
            if env_label and label == 'docker-dev':
                # docker-dev is some older master, so it must work
                # with the current master as well as with branches based on the current master.
                label = 'master'
            if label == 'master' and env_label and env_label != 'docker-dev':
                # assuming that the branch given in the label is
                # up-to-date with respect to master branch. If it is
                # not, one should rebase the branch against the
                # master.
                label = env_label
            if env_label is None:
                warnings.warn('Environment does not specify label (OMNISCIDB_DEV_LABEL is unset).'
                              ' Tests with development labels will not be run.')
            if env_label != label:
                _reason = (f'test requires version {version} with label {label},'
                           f' got {available_version} with label {env_label}')
                if message is not None:
                    _reason += f': {message}'
                pytest.skip(_reason)

            if version < available_version[:3]:
                # in the case the branch given in the label was never
                # merged, consider removing the corresponding test
                warnings.warn(f'detected test requiring {version} with out-of-date label {label}.'
                              ' Please reset test label to None.')

    # Throw an error on Travis CI if the server is not available
    if "TRAVIS" in os.environ and not available_version:
        pytest.exit(msg=reason, returncode=1)

    require_version(minimal_version)

    filename = caller_globals['__file__']
    table_name = os.path.splitext(os.path.basename(filename))[0]

    config = rbc_omnisci.get_client_config(debug=debug)
    m = rbc_omnisci.RemoteOmnisci(**config)

    if not load_test_data:
        yield m
        return

    sqltypes = ['FLOAT', 'DOUBLE', 'TINYINT', 'SMALLINT', 'INT', 'BIGINT',
                'BOOLEAN']
    arrsqltypes = [t + '[]' for t in sqltypes]
    # todo: TEXT ENCODING DICT, TEXT ENCODING NONE, TIMESTAMP, TIME,
    # DATE, DECIMAL/NUMERIC, GEOMETRY: POINT, LINESTRING, POLYGON,
    # MULTIPOLYGON, See
    # https://www.omnisci.com/docs/latest/5_datatypes.html
    colnames = ['f4', 'f8', 'i1', 'i2', 'i4', 'i8', 'b']
    table_defn = ',\n'.join('%s %s' % (n, t)
                            for t, n in zip(sqltypes, colnames))
    arrtable_defn = ',\n'.join('%s %s' % (n, t)
                               for t, n in zip(arrsqltypes, colnames))

    for suffix in suffices:
        m.sql_execute(f'DROP TABLE IF EXISTS {table_name}{suffix}')
        if 'array' in suffix:
            m.sql_execute(f'CREATE TABLE IF NOT EXISTS {table_name}{suffix} ({arrtable_defn});')
        else:
            m.sql_execute(f'CREATE TABLE IF NOT EXISTS {table_name}{suffix} ({table_defn});')

    if load_columnar:
        # fast method using load_table_columnar thrift endpoint, use for large tables
        def row_value(row, col, colname, null=False, arr=False):
            if arr:
                if null and (0 == (row + col) % 2):
                    return None
                a = [row_value(row + i, col, colname, null=null, arr=False) for i in range(row)]
                return a
            if null and (0 == (row + col) % 3):
                return None
            if colname == 'b':
                return row % 2 == 0
            return row

        for suffix in suffices:
            columns = defaultdict(list)
            for j, n in enumerate(colnames):
                for i in range(10 if '10' in suffix else 5):
                    v = row_value(i, j, n, null=('null' in suffix), arr=('array' in suffix))
                    columns[n].append(v)
            m.load_table_columnar(f'{table_name}{suffix}', **columns)

    else:
        # slow method using SQL query statements
        def row_value(row, col, colname, null=False, arr=False):
            if arr:
                if null and (0 == (row + col) % 2):
                    return 'NULL'
                a = [row_value(row + i, col, colname, null=null, arr=False) for i in range(row)]
                return '{' + ', '.join(map(str, a)) + '}'
            if null and (0 == (row + col) % 3):
                return 'NULL'
            if colname == 'b':
                return ("'true'" if row % 2 == 0 else "'false'")
            return row

        for i in range(10):
            if i < 5:
                for suffix in suffices:
                    if suffix == '':
                        table_row = ', '.join(str(row_value(i, j, n))
                                              for j, n in enumerate(colnames))
                    elif suffix == 'null':
                        table_row = ', '.join(str(row_value(i, j, n, null=True))
                                              for j, n in enumerate(colnames))
                    elif suffix == 'array':
                        table_row = ', '.join(str(row_value(i, j, n, arr=True))
                                              for j, n in enumerate(colnames))
                    elif suffix == 'arraynull':
                        table_row = ', '.join(str(row_value(i, j, n, null=True, arr=True))
                                              for j, n in enumerate(colnames))
                    else:
                        continue
                    m.sql_execute(f'INSERT INTO {table_name}{suffix} VALUES ({table_row})')
            if i < 10 and '10' in suffices:
                table_row = ', '.join(str(row_value(i, j, n)) for j, n in enumerate(colnames))
                m.sql_execute(f'INSERT INTO {table_name}10 VALUES ({table_row})')

    m.table_name = table_name
    m.require_version = require_version
    yield m
    for suffix in suffices:
        m.sql_execute(f'DROP TABLE IF EXISTS {table_name}{suffix}')



import pytest
from rbc.tests import omnisci_fixture

scalar_types = ['float64', 'int64', 'float32', 'int32', 'int16', 'int8']
table_columns_map = dict(int64='i8', int32='i4', int16='i2', int8='i1',
                         float64='f8', float32='f4')
sql_type_map = dict(int64='BIGINT', int32='INT', int16='SMALLINT', int8='TINYINT',
                    float64='DOUBLE', float32='FLOAT')


@pytest.fixture(scope='module')
def omnisci():

    for o in omnisci_fixture(globals(), minimal_version=(5, 6)):
        define(o)
        yield o


def define(omnisci):

    @omnisci('int32 columns_sum1(Cursor<Column<int64>, ColumnList<T>>,'
             ' RowMultiplier, OutputColumn<T>)',
             T=scalar_types, devices=['cpu'])
    def columns_sum1(rowid, lst, m, out):
        for i in range(lst.size):
            out[i] *= 0
            for j in range(lst.length):
                out[i] += lst.ptrs[j][i]
        return lst.size

    @omnisci('int32 columns_sum2(Cursor<Column<int64>, ColumnList<T>>,'
             ' RowMultiplier, OutputColumn<T>)',
             T=scalar_types, devices=['cpu'])
    def columns_sum2(rowid, lst, m, out):
        for j in range(lst.length):
            col = lst[j]  # equivalent to lst.ptrs[j]
            if j == 0:
                for i in range(lst.size):
                    out[i] = col[i]
            else:
                for i in range(lst.size):
                    out[i] += col[i]
        return lst.size

    @omnisci('int32 columns_sum3(Cursor<Column<int64>, ColumnList<T>>,'
             ' RowMultiplier, OutputColumn<T>)',
             T=scalar_types, devices=['cpu'])
    def columns_sum3(rowid, lst, m, out):
        for j in range(lst.ncols):
            col = lst[j]
            for i in range(lst.nrows):
                out[i] = col[i] if j == 0 else out[i] + col[i]
        return lst.nrows


@pytest.mark.parametrize("variant", ['1', "2", "3"])
@pytest.mark.parametrize("T", scalar_types)
def test_columns_sum(omnisci, T, variant):
    v = table_columns_map[T]
    c = sql_type_map[T]

    query = f'select rowid, ({v} + {v} + 1) from {omnisci.table_name} ORDER BY rowid'
    _, result = omnisci.sql_execute(query)
    expected = list(zip(*result))[1]

    # using `.. ORDER BY rowid LIMIT ALL` will crash omniscidb server
    query = (f'select * from table(columns_sum{variant}(cursor('
             f'select rowid, {v}, CAST({v}+1 as {c}) '
             f'from {omnisci.table_name} ORDER BY rowid LIMIT 1000), 1))')
    _, result = omnisci.sql_execute(query)
    actual = list(zip(*result))[0]

    assert expected == actual



import sys
import atexit
import pytest
from rbc.tests import omnisci_fixture
from rbc.remotejit import RemoteJIT
from rbc.externals import cmath
from rbc.typesystem import Type
import numpy as np
import math


@pytest.fixture(scope="module")
def rjit(request):
    local = False
    rjit = RemoteJIT(debug=not True, local=local)
    if not local:
        rjit.start_server(background=True)
        request.addfinalizer(rjit.stop_server)
        atexit.register(rjit.stop_server)
    define(rjit)
    return rjit


@pytest.fixture(scope="module")
def ljit(request):
    ljit = RemoteJIT(debug=not True, local=True)
    define(ljit)
    return ljit


@pytest.fixture(scope="module")
def omnisci():

    for o in omnisci_fixture(globals()):
        define(o)
        yield o


cmath_funcs = (
    # Trigonometric
    ("cos", "float64 cos(float64)|CPU"),
    ("sin", "float64 sin(float64)|CPU"),
    ("tan", "float64 tan(float64)|CPU"),
    ("acos", "float64 acos(float64)|CPU"),
    ("asin", "float64 asin(float64)|CPU"),
    ("atan", "float64 atan(float64)|CPU"),
    ("atan2", "float64 atan2(float64, float64)|CPU"),
    # Hyperbolic
    ("cosh", "float64 cosh(float64)|CPU"),
    ("sinh", "float64 sinh(float64)|CPU"),
    ("tanh", "float64 tanh(float64)|CPU"),
    ("acosh", "float64 acosh(float64)|CPU"),
    ("asinh", "float64 asinh(float64)|CPU"),
    ("atanh", "float64 atanh(float64)|CPU"),
    # Exponential and logarithmic functions
    ("exp", "float64 exp(float64)|CPU"),
    ("frexp", "float64 frexp(float64)|CPU"),
    ("ldexp", "float64 ldexp(float64, int64)|CPU"),
    ("log", "float64 log(float64)|CPU"),
    ("log10", "float64 log10(float64)|CPU"),
    ("modf", "float64 modf(float64, float64*)|CPU"),
    ("exp2", "float64 exp2(float64)|CPU"),
    ("expm1", "float64 expm1(float64)|CPU"),
    ("ilogb", "float64 ilogb(float64)|CPU"),
    ("log1p", "float64 log1p(float64)|CPU"),
    ("log2", "float64 log2(float64)|CPU"),
    ("logb", "float64 logb(float64)|CPU"),
    # power functions
    ("pow", "float64 pow(float64, float64)|CPU"),
    ("sqrt", "float64 sqrt(float64)|CPU"),
    ("cbrt", "float64 cbrt(float64)|CPU"),
    ("hypot", "float64 hypot(float64, float64)|CPU"),
    # error and gamma functions
    ("erf", "float64 erf(float64)|CPU"),
    ("erfc", "float64 erfc(float64)|CPU"),
    ("tgamma", "float64 tgamma(float64)|CPU"),
    ("lgamma", "float64 lgamma(float64)|CPU"),
    # Rounding
    ("ceil", "float64 ceil(float64)|CPU"),
    ("floor", "float64 floor(float64)|CPU"),
    ("fmod", "float64 fmod(float64, float64)|CPU"),
    ("trunc", "float64 trunc(float64)|CPU"),
    ("round", "float64 round(float64)|CPU"),
    ("lround", "int64 lround(float64)|CPU"),
    ("llround", "int64 llround(float64)|CPU"),
    ("rint", "float64 rint(float64)|CPU"),
    ("lrint", "int64 lrint(float64)|CPU"),
    ("llrint", "int64 llrint(float64)|CPU"),
    ("nearbyint", "float64 nearbyint(float64)|CPU"),
    ("remainder", "float64 remainder(float64, float64)|CPU"),
    # Floating-point manipulation
    ("copysign", "float64 copysign(float64, float64)|CPU"),
    ("nan", "float64 nan(const char*)|CPU"),
    ("nextafter", "float64 nextafter(float64, float64)|CPU"),
    ("nexttoward", "float64 nexttoward(float64, float64)|CPU"),
    # Minimum, maximum, difference functions
    ("fdim", "float64 fdim(float64, float64)|CPU"),
    ("fmax", "float64 fmax(float64, float64)|CPU"),
    ("fmin", "float64 fmin(float64, float64)|CPU"),
    # Other functions
    ("fabs", "float64 fabs(float64)|CPU"),
    ("abs", "int64 abs(int64)|CPU"),
    ("fma", "float64 fma(float64, float64, float64)|CPU"),
)


funcs = {}


def _get_query_func(fname):
    PREFIX = "test_cmath"
    return f"{PREFIX}_{fname}"


def define(jit):
    global funcs

    def inner(cmath_func, query_func, signature):
        cmath_fn = getattr(cmath, cmath_func)
        t = Type.fromstring(signature)
        retty = str(t[0])
        argtypes = tuple(map(str, t[1]))
        arity = len(argtypes)

        # define callable
        if arity == 1:
            def fn(a):
                return cmath_fn(a)

        elif arity == 2:
            def fn(a, b):
                return cmath_fn(a, b)

        else:
            def fn(a, b, c):
                return cmath_fn(a, b, c)

        fn.__name__ = query_func
        fn = jit(f"{retty}({', '.join(argtypes)})", devices=["cpu"])(fn)
        return fn

    blocklist = ("frexp", "nan")

    for fname, signature in cmath_funcs:
        if fname not in blocklist:
            query_func = _get_query_func(fname)
            fn = inner(fname, query_func, signature)
            funcs[(jit, query_func)] = fn


def _get_pyfunc(fname):
    remap = {
        "acos": "arccos",
        "asin": "arcsin",
        "atan": "arctan",
        "atan2": "arctan2",
        "acosh": "arccosh",
        "asinh": "arcsinh",
        "atanh": "arctanh",
        "pow": "power",
        "tgamma": "gamma",
        "lround": "round",
        "llround": "round",
        "lrint": "round",
        "llrint": "round",
        "nearbyint": "round",
        "nexttoward": "nextafter",
        "fdim": "subtract",
    }

    if fname == "fma":
        return lambda a, b, c: a*b + c
    if fname == "ilogb":
        return lambda a: np.trunc(np.log2(a))
    if fname == "logb":
        return np.log2

    mod = math if fname in ["erf", "erfc", "tgamma", "lgamma"] else np
    fn = getattr(mod, remap.get(fname, fname))
    return fn


@pytest.mark.parametrize("fname,sig", cmath_funcs, ids=[item[0] for item in cmath_funcs])
def test_external_cmath_omnisci(omnisci, fname, sig):

    if fname in ["logb", "ilogb", "modf"]:
        pytest.skip(f"cmath function {fname} not supported")

    if fname in ["frexp", "nan"]:
        pytest.xfail(f"cmath function {fname} crashes omniscidb server")

    if fname in ["remainder"]:
        pytest.xfail(f"cmath.{fname} wrong output!")

    table = omnisci.table_name
    query_func = _get_query_func(fname)
    pyfunc = _get_pyfunc(fname)

    if fname in ["acos", "asin", "atan"]:
        query = f"SELECT f8/10.0, {query_func}(f8/10.0) from {table}"
    elif fname in ["atan2"]:
        query = f"SELECT f8/10.0, f8/8.0, {query_func}(f8/10.0, f8/8.0) FROM {table}"
    elif fname in [
        "pow",
        "hypot",
        "fmod",
        "remainder",
        "nextafter",
        "nexttoward",
        "fdim",
        "fmax",
        "fmin",
    ]:
        query = f"SELECT f8+10.0, f8+1.0, {query_func}(f8+10.0, f8+1.0) FROM {table}"
    elif fname == "copysign":
        query = f"SELECT f8, -1*f8, {query_func}(f8, -1*f8) FROM {table}"
    elif fname == "fma":
        query = f"SELECT f8, f8, f8, {query_func}(f8, f8, f8) FROM {table}"
    elif fname == "ldexp":
        query = f"SELECT f8+1.0, 2, {query_func}(f8+1.0, 2) FROM {table}"
    elif fname == "atanh":
        query = f"SELECT f8/8.0, {query_func}(f8/8.0) from {table}"
    elif fname == "abs":
        query = f"SELECT -1*i8, {query_func}(-1*i8) from {table}"
    else:
        query = f"SELECT f8+10.0, {query_func}(f8+10.0) from {table}"

    _, result = omnisci.sql_execute(query)

    for values in result:
        if len(values) == 3:
            a, b, r = values
            assert np.isclose(r, pyfunc(a, b)), fname
        elif len(values) == 4:
            a, b, c, r = values
            assert np.isclose(r, pyfunc(a, b, c)), fname
        else:
            a, r = values
            assert np.isclose(r, pyfunc(a)), fname


@pytest.fixture(scope="module")
def input_data():
    return {
        'f8': np.arange(5, dtype=np.float64),
        'i8': np.arange(5, dtype=np.int64)
    }


@pytest.mark.parametrize("location", ['local', 'remote'])
@pytest.mark.parametrize("fname,sig", cmath_funcs, ids=[item[0] for item in cmath_funcs])
def test_external_cmath_remotejit(input_data, location, ljit, rjit, fname, sig):

    if fname in ["modf", "frexp", "nan"]:
        pytest.skip(f"cmath function {fname} requires a pointer argument")

    if fname in ["remainder", "ilogb", "logb"]:
        pytest.xfail(f"cmath function {fname} returns the wrong value")

    if fname in ["nexttoward"] and sys.platform == "darwin":
        pytest.xfail(f"{fname} fails on {sys.platform}")

    jit = rjit if location == 'remote' else ljit
    query_func = _get_query_func(fname)
    fn = funcs.get((jit, query_func))

    i8 = input_data['i8']
    f8 = input_data['f8']

    pyfunc = _get_pyfunc(fname)

    if fname in ["acos", "asin", "atan"]:
        args = (f8/10.0,)
    elif fname in ["logb"]:
        args = ((f8+1)*10.0,)
    elif fname in ["atan2"]:
        args = (f8/10.0, f8/8.0)
    elif fname in ["pow", "hypot", "fmod", "remainder", "nextafter",
                   "nexttoward", "fdim", "fmax", "fmin"]:
        args = (f8+10.0, f8+1.0)
    elif fname == "copysign":
        args = (f8, -1*f8)
    elif fname == "fma":
        args = (f8, f8, f8)
    elif fname == "ldexp":
        args = f8+1.0, np.full_like(f8, 2, dtype=np.int32)
    elif fname == "atanh":
        args = (f8/8.0,)
    elif fname == "abs":
        args = (-1*i8,)
    else:
        args = (f8 + 10.0,)

    result = list(map(lambda inputs: fn(*inputs), zip(*args)))

    for values in zip(*args, result):
        if len(values) == 3:
            a, b, r = values
            assert np.isclose(r, pyfunc(a, b)), fname
        elif len(values) == 4:
            a, b, c, r = values
            assert np.isclose(r, a * b + c), fname
        else:
            a, r = values
            assert np.isclose(r, pyfunc(a)), fname



import pytest
import numpy as np
from rbc.tests import omnisci_fixture
from rbc.omnisci_backend import Array


rbc_omnisci = pytest.importorskip('rbc.omniscidb')
available_version, reason = rbc_omnisci.is_available()
pytestmark = pytest.mark.skipif(not available_version, reason=reason)


NUMERIC_TYPES = ['int8', 'int16', 'int32', 'int64', 'float32', 'float64']


@pytest.fixture(scope='module')
def omnisci():
    for o in omnisci_fixture(globals()):
        define(o)
        yield o


ndarray_methods = [
    ('fill', (5, 4), [4.0, 4.0, 4.0, 4.0, 4.0]),
    ('max', (5, 4.0), 4.0),
    ('max_empty_int8', (0, ), np.iinfo(np.int8).min + 1),
    ('max_empty_int16', (0, ), np.iinfo(np.int16).min + 1),
    ('max_empty_int32', (0, ), np.iinfo(np.int32).min + 1),
    ('max_empty_int64', (0, ), np.iinfo(np.int64).min + 1),
    ('max_empty_float32', (0, ), np.finfo(np.float32).min),
    ('max_empty_float64', (0, ), np.finfo(np.float64).min),
    ('mean', (5, 2), 2.0),
    ('mean', (5, 2.0), 2.0),
    ('mean_empty_int8', (0, ), np.nan),
    ('mean_empty_int16', (0, ), np.nan),
    ('mean_empty_int32', (0, ), np.nan),
    ('mean_empty_int64', (0, ), np.nan),
    ('mean_empty_float32', (0, ), np.nan),
    ('mean_empty_float64', (0, ), np.nan),
    ('min', (5, 4.0), 4.0),
    ('min_empty_int8', (0, ), np.iinfo(np.int8).max),
    ('min_empty_int16', (0, ), np.iinfo(np.int16).max),
    ('min_empty_int32', (0, ), np.iinfo(np.int32).max),
    ('min_empty_int64', (0, ), np.iinfo(np.int64).max),
    ('min_empty_float32', (0, ), np.finfo(np.float32).max),
    ('min_empty_float64', (0, ), np.finfo(np.float64).max),
    ('sum', (5, 2.0), 10.0),
    ('prod', (5, 3.0), 243.0),
]


def define(omnisci):

    @omnisci('double[](int64, double)')
    def ndarray_fill(size, v):
        a = Array(size, 'double')
        a.fill(v)
        return a

    @omnisci('double(int64, double)')
    def ndarray_max(size, v):
        a = Array(size, 'double')
        a.fill(v)
        return a.max()

    for retty in NUMERIC_TYPES:
        for op in ('min', 'max', 'mean'):
            fn_name = f'ndarray_{op}_empty_{retty}'
            fn = (f'def {fn_name}(size):\n'
                  f'    a = Array(size, "{retty}")\n'
                  f'    return a.{op}()\n')
            exec(fn)
            fn = locals()[fn_name]
            if op == 'mean':
                omnisci('float64(int32)')(fn)
            else:
                omnisci(f'{retty}(int32)')(fn)

    @omnisci('double(int64, double)')
    def ndarray_mean(size, v):
        a = Array(size, 'double')
        a.fill(v)
        return a.mean()

    @omnisci('double(int64, double)')
    def ndarray_min(size, v):
        a = Array(size, 'double')
        a.fill(v)
        return a.min()

    @omnisci('double(int64, double)')
    def ndarray_sum(size, v):
        a = Array(size, 'double')
        a.fill(v)
        return a.sum()

    @omnisci('double(int64, double)')
    def ndarray_prod(size, v):
        a = Array(size, 'double')
        a.fill(v)
        return a.prod()


@pytest.mark.parametrize("method, args, expected", ndarray_methods,
                         ids=[item[0] for item in ndarray_methods])
def test_ndarray_methods(omnisci, method, args, expected):
    query_args = ', '.join(map(str, args))
    query = f'SELECT ndarray_{method}({query_args})'

    _, result = omnisci.sql_execute(query)
    out = list(result)[0]

    if method == 'fill':
        assert np.array_equal(expected, out[0]), 'ndarray_' + method
    else:
        assert np.isclose(expected, out, equal_nan=True), \
            'ndarray_' + method



class TracingAllocatorError(Exception):
    pass


class InvalidFreeError(TracingAllocatorError):
    pass


class MemoryLeakError(TracingAllocatorError):

    def __init__(self, leaks):
        lines = [f'Found {len(leaks)} memory leaks:']
        for addr, seq in leaks:
            lines.append(f'    {addr} (seq = {seq})')
        message = '\n'.join(lines)
        super().__init__(message)
        self.leaks = leaks



"""Implement Omnisci Array type support
"""

__all__ = ['ArrayPointer', 'Array', 'OmnisciArrayType']

from rbc import typesystem, errors
from .omnisci_buffer import (BufferPointer, Buffer,
                             OmnisciBufferType,
                             omnisci_buffer_constructor)
from numba.core import extending, types


class OmnisciArrayType(OmnisciBufferType):
    """Omnisci Array type for RBC typesystem.

    :code:`Omnisci Array<T>` is defined as follows (using C++ syntax)::

      template<typename T>
      struct Array {
        T* ptr;
        size_t sz;
        bool is_null;
      }
    """
    @property
    def buffer_extra_members(self):
        return ('bool is_null',)


ArrayPointer = BufferPointer


class Array(Buffer):
    pass


@extending.lower_builtin(Array, types.Integer, types.StringLiteral)
@extending.lower_builtin(Array, types.Integer, types.NumberClass)
def omnisci_array_constructor(context, builder, sig, args):
    return omnisci_buffer_constructor(context, builder, sig, args)


@extending.type_callable(Array)
def type_omnisci_array(context):
    def typer(size, dtype):
        if isinstance(dtype, types.StringLiteral):
            element_type = typesystem.Type.fromstring(dtype.literal_value)
        elif isinstance(dtype, types.NumberClass):
            element_type = typesystem.Type.fromobject(dtype)
        else:
            raise errors.NumbaNotImplementedError(repr(dtype))
        return OmnisciArrayType((element_type,)).tonumba()
    return typer



"""https://en.cppreference.com/w/c/io
"""

from rbc import irutils
from llvmlite import ir
from rbc.targetinfo import TargetInfo
from numba.core import cgutils, extending
from numba.core import types as nb_types
from rbc.errors import NumbaTypeError  # some errors are available for Numba >= 0.55

int32_t = ir.IntType(32)


def _cg_fflush(builder):
    int8_t = ir.IntType(8)
    fflush_fnty = ir.FunctionType(int32_t, [int8_t.as_pointer()])
    fflush_fn = irutils.get_or_insert_function(builder.module, fflush_fnty, name="fflush")

    builder.call(fflush_fn, [int8_t.as_pointer()(None)])


@extending.intrinsic
def fflush(typingctx):
    """``fflush`` that can be called from Numba jit-decorated functions.

    .. note::
        ``fflush`` is available only for CPU target.
    """
    sig = nb_types.void(nb_types.void)

    def codegen(context, builder, signature, args):
        target_info = TargetInfo()
        if target_info.is_cpu:
            _cg_fflush(builder)

    return sig, codegen


@extending.intrinsic
def printf(typingctx, format_type, *args):
    """``printf`` that can be called from Numba jit-decorated functions.

    .. note::
        ``printf`` is available only for CPU target.
    """

    if isinstance(format_type, nb_types.StringLiteral):
        sig = nb_types.void(format_type, nb_types.BaseTuple.from_types(args))

        def codegen(context, builder, signature, args):
            target_info = TargetInfo()
            if target_info.is_cpu:
                cgutils.printf(builder, format_type.literal_value, *args[1:])
                _cg_fflush(builder)

        return sig, codegen

    else:
        raise NumbaTypeError(f"expected StringLiteral but got {type(format_type).__name__}")



"""Implement Buffer type as a base class to Omnisci Array and Column types.

Omnisci Buffer represents the following structure:

  template<typename T>
  struct Buffer {
    T* ptr;
    size_t sz;
    ...
  }

that is, a structure that has at least two members where the first is
a pointer to some data type and the second is the size of the buffer.

This module implements the support for the following Python operators:

  len
  __getitem__
  __setitem__

to provide a minimal functionality for accessing and manipulating the
Omnisci buffer objects from UDF/UDTFs.
"""


import operator
from collections import defaultdict
from .omnisci_metatype import OmnisciMetaType
from llvmlite import ir
import numpy as np
from rbc import typesystem, irutils
from rbc.targetinfo import TargetInfo
from llvmlite import ir as llvm_ir
from numba.core import datamodel, cgutils, extending, types, imputils
from rbc.errors import UnsupportedError

int8_t = ir.IntType(8)
int32_t = ir.IntType(32)
int64_t = ir.IntType(64)
void_t = ir.VoidType()
fp32 = ir.FloatType()
fp64 = ir.DoubleType()


class OmnisciBufferType(typesystem.Type):
    """Typesystem type class for Omnisci buffer structures.
    """
    # When True, buffer type arguments are passed by value to
    # functions [not recommended].
    @property
    def pass_by_value(self):
        return False

    @classmethod
    def preprocess_args(cls, args):
        assert len(args) == 1, args
        assert len(args[0]) == 1, args
        element_type = args[0][0]
        if not isinstance(element_type, typesystem.Type):
            element_type = typesystem.Type.fromobject(element_type)
        return ((element_type,),)

    @property
    def element_type(self):
        return self[0][0]

    @property
    def buffer_extra_members(self):
        return ()

    def tonumba(self, bool_is_int8=None):
        ptr_t = typesystem.Type(self.element_type, '*', name='ptr')
        size_t = typesystem.Type.fromstring('size_t sz')
        extra_members = tuple(map(typesystem.Type.fromobject, self.buffer_extra_members))
        buffer_type = typesystem.Type(
            ptr_t,
            size_t,
            *extra_members
        )
        buffer_type._params['NumbaType'] = BufferType
        buffer_type._params['NumbaPointerType'] = BufferPointer
        numba_type = buffer_type.tonumba(bool_is_int8=True)
        if self.pass_by_value:
            return numba_type
        return BufferPointer(numba_type)


class BufferType(types.Type):
    """Numba type class for Omnisci buffer structures.
    """

    @property
    def eltype(self):
        """
        Return buffer element dtype.
        """
        return self.members[0].dtype


class BufferPointer(types.IterableType):
    """Numba type class for pointers to Omnisci buffer structures.

    We are not deriving from CPointer because BufferPointer getitem is
    used to access the data stored in Buffer ptr member.
    """
    mutable = True
    return_as_first_argument = True

    def __init__(self, dtype):
        self.dtype = dtype    # struct dtype
        self.eltype = dtype.eltype  # buffer element dtype
        name = "%s[%s]*" % (type(self).__name__, dtype)
        super().__init__(name)

    @property
    def key(self):
        return self.dtype

    @property
    def iterator_type(self):
        return BufferPointerIteratorType(self)


class BufferPointerIteratorType(types.SimpleIteratorType):

    def __init__(self, buffer_type):
        name = f"iter_buffer({buffer_type})"
        self.buffer_type = buffer_type
        super().__init__(name, self.buffer_type.eltype)


@datamodel.register_default(BufferPointerIteratorType)
class BufferPointerIteratorModel(datamodel.StructModel):
    def __init__(self, dmm, fe_type):
        members = [('index', types.EphemeralPointer(types.uintp)),
                   ('buffer', fe_type.buffer_type)]
        super(BufferPointerIteratorModel, self).__init__(dmm, fe_type, members)


class BufferMeta(OmnisciMetaType):
    pass


class Buffer(object, metaclass=BufferMeta):
    """Represents Omnisci Buffer that can be constructed within UDF/UDTFs.
    """


@datamodel.register_default(BufferPointer)
class BufferPointerModel(datamodel.models.PointerModel):
    """Base class for Omnisci buffer pointer models.

    Subclasses should register the model for the corresponding
    BufferPointer subclass, for instance::

      @datamodel.register_default(ArrayPointer)
      class ArrayPointerModel(BufferPointerModel):
          pass
    """


builder_buffers = defaultdict(list)


def omnisci_buffer_constructor(context, builder, sig, args):
    """

    Usage:

      extending.lower_builtin(MyBuffer, numba.types.Integer, ...)(omnisci_buffer_constructor)

    will enable creating MyBuffer instance from a Omnisci UDF/UDTF definition:

      b = MyBuffer(<size>, ...)

    """
    target_info = TargetInfo()
    try:
        alloc_fn_name = target_info.info['fn_allocate_varlen_buffer']
    except KeyError as msg:
        raise UnsupportedError(f'{target_info} does not provide {msg}')

    ptr_type, sz_type = sig.return_type.dtype.members[:2]
    if len(sig.return_type.dtype.members) > 2:
        assert len(sig.return_type.dtype.members) == 3
        null_type = sig.return_type.dtype.members[2]
    else:
        null_type = None
    assert isinstance(args[0].type, ir.IntType), (args[0].type)
    element_count = builder.zext(args[0], int64_t)
    element_size = int64_t(ptr_type.dtype.bitwidth // 8)

    alloc_fnty = ir.FunctionType(int8_t.as_pointer(), [int64_t, int64_t])

    alloc_fn = irutils.get_or_insert_function(builder.module, alloc_fnty, alloc_fn_name)
    ptr8 = builder.call(alloc_fn, [element_count, element_size])
    # remember possible temporary allocations so that when leaving a
    # UDF/UDTF, these will be deallocated, see omnisci_pipeline.py.
    builder_buffers[builder].append(ptr8)
    ptr = builder.bitcast(ptr8, context.get_value_type(ptr_type))

    fa = cgutils.create_struct_proxy(sig.return_type.dtype)(context, builder)
    fa.ptr = ptr                  # T*
    fa.sz = element_count         # size_t
    if null_type is not None:
        is_zero = builder.icmp_signed('==', element_count, int64_t(0))
        with builder.if_else(is_zero) as (then, orelse):
            with then:
                is_null = context.get_value_type(null_type)(1)
            with orelse:
                is_null = context.get_value_type(null_type)(0)
        fa.is_null = is_null      # int8_t
    return fa._getpointer()


@extending.intrinsic
def free_all_other_buffers(typingctx, value_to_keep_alive):
    """
    Black magic function which automatically frees all the buffers which were
    allocated in the function apart the given one.

    value_to_keep_alive can be of any type:
      - if it's of a Buffer type, it will be kept alive and not freed
      - if it's of any other type, all buffers will be freed unconditionally

    The end user should never call this function explicitly: it is
    automatically inserted by omnisci_pipeline.AutoFreeBuffers.
    """

    sig = types.void(value_to_keep_alive)

    def codegen(context, builder, signature, args):
        buffers = builder_buffers[builder]

        # TODO: using stdlib `free` that works only for CPU. For CUDA
        # devices, we need to use omniscidb provided deallocator.
        target_info = TargetInfo()
        try:
            free_buffer_fn_name = target_info.info['fn_free_buffer']
        except KeyError as msg:
            raise UnsupportedError(f'{target_info} does not provide {msg}')
        free_buffer_fnty = llvm_ir.FunctionType(void_t, [int8_t.as_pointer()])
        free_buffer_fn = irutils.get_or_insert_function(
            builder.module, free_buffer_fnty, free_buffer_fn_name)

        if isinstance(value_to_keep_alive, BufferPointer):
            # free all the buffers apart value_to_keep_alive
            [keep_alive] = args
            keep_alive_ptr = builder.load(builder.gep(keep_alive, [int32_t(0), int32_t(0)]))
            keep_alive_ptr = builder.bitcast(keep_alive_ptr, int8_t.as_pointer())
            for ptr8 in buffers:
                with builder.if_then(builder.icmp_signed('!=', keep_alive_ptr, ptr8)):
                    builder.call(free_buffer_fn, [ptr8])
        else:
            # free all the buffers unconditionally
            for ptr8 in buffers:
                builder.call(free_buffer_fn, [ptr8])

        del builder_buffers[builder]

    return sig, codegen


@extending.intrinsic
def free_buffer(typingctx, buf):
    """
    Free a buffer
    """
    sig = types.void(buf)
    assert isinstance(buf, BufferPointer)

    def codegen(context, builder, signature, args):
        # TODO: using stdlib `free` that works only for CPU. For CUDA
        # devices, we need to use omniscidb provided deallocator.
        target_info = TargetInfo()
        free_buffer_fn_name = target_info.info['fn_free_buffer']
        free_buffer_fnty = llvm_ir.FunctionType(void_t, [int8_t.as_pointer()])
        free_buffer_fn = irutils.get_or_insert_function(
            builder.module, free_buffer_fnty, free_buffer_fn_name)

        [buf] = args
        buf_ptr = builder.load(builder.gep(buf, [int32_t(0), int32_t(0)]))  # buf.ptr
        buf_ptr = builder.bitcast(buf_ptr, int8_t.as_pointer())
        builder.call(free_buffer_fn, [buf_ptr])

    return sig, codegen


@extending.intrinsic
def omnisci_buffer_ptr_get_ptr_(typingctx, data):
    eltype = data.eltype
    ptrtype = types.CPointer(eltype)
    sig = ptrtype(data)

    def codegen(context, builder, signature, args):
        data,  = args
        rawptr = cgutils.alloca_once_value(builder, value=data)
        struct = builder.load(builder.gep(rawptr,
                                          [int32_t(0)]))
        return builder.load(builder.gep(struct, [int32_t(0), int32_t(0)]))

    return sig, codegen


@extending.intrinsic
def omnisci_buffer_get_ptr_(typingctx, data):
    eltype = data.eltype
    ptrtype = types.CPointer(eltype)
    sig = ptrtype(data)

    def codegen(context, builder, signature, args):
        data, = args
        assert data.opname == 'load'
        struct = data.operands[0]
        return builder.load(builder.gep(struct, [int32_t(0), int32_t(0)]))

    return sig, codegen


@extending.intrinsic
def omnisci_buffer_ptr_item_get_ptr_(typingctx, data, index):
    eltype = data.eltype
    ptrtype = types.CPointer(eltype)
    sig = ptrtype(data, index)

    def codegen(context, builder, signature, args):
        data, index = args
        rawptr = cgutils.alloca_once_value(builder, value=data)
        struct = builder.load(builder.gep(rawptr, [int32_t(0)]))
        ptr = builder.load(builder.gep(struct, [int32_t(0), int32_t(0)]))
        return builder.gep(ptr, [index])

    return sig, codegen


@extending.overload_method(BufferPointer, 'ptr')
def omnisci_buffer_get_ptr(x, index=None):
    if isinstance(x, BufferPointer):
        if cgutils.is_nonelike(index):
            def impl(x, index=None):
                return omnisci_buffer_ptr_get_ptr_(x)
        else:
            def impl(x, index=None):
                return omnisci_buffer_ptr_item_get_ptr_(x, index)
        return impl
    if isinstance(x, BufferType):
        if cgutils.is_nonelike(index):
            def impl(x, index=None):
                return omnisci_buffer_get_ptr_(x)
        else:
            raise NotImplementedError(f'omnisci_buffer_item_get_ptr_({x}, {index})')
        return impl


@extending.intrinsic
def omnisci_buffer_ptr_len_(typingctx, data):
    sig = types.int64(data)

    def codegen(context, builder, signature, args):
        data, = args
        rawptr = cgutils.alloca_once_value(builder, value=data)
        struct = builder.load(builder.gep(rawptr,
                                          [int32_t(0)]))
        return builder.load(builder.gep(
            struct, [int32_t(0), int32_t(1)]))
    return sig, codegen


@extending.intrinsic
def omnisci_buffer_len_(typingctx, data):
    sig = types.int64(data)

    def codegen(context, builder, signature, args):
        data, = args
        return irutils.get_member_value(builder, data, 1)

    return sig, codegen


@extending.overload(len)
def omnisci_buffer_len(x):
    if isinstance(x, BufferPointer):
        return lambda x: omnisci_buffer_ptr_len_(x)
    if isinstance(x, BufferType):
        return lambda x: omnisci_buffer_len_(x)


@extending.intrinsic
def omnisci_buffer_ptr_getitem_(typingctx, data, index):
    sig = data.eltype(data, index)

    def codegen(context, builder, signature, args):
        data, index = args
        rawptr = cgutils.alloca_once_value(builder, value=data)
        buf = builder.load(builder.gep(rawptr, [int32_t(0)]))
        ptr = builder.load(builder.gep(
            buf, [int32_t(0), int32_t(0)]))
        res = builder.load(builder.gep(ptr, [index]))

        return res
    return sig, codegen


@extending.intrinsic
def omnisci_buffer_getitem_(typingctx, data, index):
    eltype = data.eltype
    sig = eltype(data, index)

    def codegen(context, builder, signature, args):
        data, index = args
        ptr = irutils.get_member_value(builder, data, 0)
        res = builder.load(builder.gep(ptr, [index]))

        return res
    return sig, codegen


@extending.overload(operator.getitem)
def omnisci_buffer_getitem(x, i):
    if isinstance(x, BufferPointer):
        return lambda x, i: omnisci_buffer_ptr_getitem_(x, i)
    if isinstance(x, BufferType):
        return lambda x, i: omnisci_buffer_getitem_(x, i)


# [rbc issue-197] Numba promotes operations like
# int32(a) + int32(b) to int64
def truncate_or_extend(builder, nb_value, eltype, value, buf_typ):
    # buf[pos] = val

    if isinstance(nb_value, types.Integer):  # Integer
        if eltype.bitwidth < nb_value.bitwidth:
            return builder.trunc(value, buf_typ)  # truncate
        elif eltype.bitwidth > nb_value.bitwidth:
            is_signed = nb_value.signed
            return builder.sext(value, buf_typ) if is_signed else \
                builder.zext(value, buf_typ)  # extend
    elif isinstance(nb_value, types.Float):  # Floating-point
        if eltype.bitwidth < nb_value.bitwidth:
            return builder.fptrunc(value, buf_typ)  # truncate
        elif eltype.bitwidth > nb_value.bitwidth:
            return builder.fpext(value, buf_typ)  # extend
    elif isinstance(nb_value, types.Boolean):
        if buf_typ.width < value.type.width:
            return builder.trunc(value, buf_typ)
        elif buf_typ.width > value.type.width:
            return builder.zext(value, buf_typ)

    return value


@extending.intrinsic
def omnisci_buffer_ptr_setitem_(typingctx, data, index, value):
    sig = types.none(data, index, value)

    eltype = data.eltype
    nb_value = value

    def codegen(context, builder, signature, args):
        zero = int32_t(0)

        data, index, value = args

        rawptr = cgutils.alloca_once_value(builder, value=data)
        ptr = builder.load(rawptr)

        buf = builder.load(builder.gep(ptr, [zero, zero]))
        value = truncate_or_extend(builder, nb_value, eltype, value, buf.type.pointee)
        builder.store(value, builder.gep(buf, [index]))

    return sig, codegen


@extending.intrinsic
def omnisci_buffer_setitem_(typingctx, data, index, value):
    sig = types.none(data, index, value)

    eltype = data.members[0].dtype
    nb_value = value

    def codegen(context, builder, signature, args):
        data, index, value = args
        ptr = irutils.get_member_value(builder, data, 0)
        value = truncate_or_extend(builder, nb_value, eltype, value, ptr.type.pointee)

        builder.store(value, builder.gep(ptr, [index]))

    return sig, codegen


@extending.overload(operator.setitem)
def omnisci_buffer_setitem(a, i, v):
    if isinstance(a, BufferPointer):
        return lambda a, i, v: omnisci_buffer_ptr_setitem_(a, i, v)
    if isinstance(a, BufferType):
        return lambda a, i, v: omnisci_buffer_setitem_(a, i, v)


@extending.intrinsic
def omnisci_buffer_is_null_(typingctx, data):
    sig = types.int8(data)

    def codegen(context, builder, sig, args):
        rawptr = cgutils.alloca_once_value(builder, value=args[0])
        ptr = builder.load(rawptr)
        return builder.load(builder.gep(ptr, [int32_t(0), int32_t(2)]))

    return sig, codegen


@extending.intrinsic
def omnisci_buffer_set_null_(typingctx, data):
    sig = types.none(data)

    def codegen(context, builder, sig, args):
        rawptr = cgutils.alloca_once_value(builder, value=args[0])
        ptr = builder.load(rawptr)
        builder.store(int8_t(1), builder.gep(ptr, [int32_t(0), int32_t(2)]))

    return sig, codegen


@extending.intrinsic
def omnisci_buffer_idx_is_null_(typingctx, col_var, row_idx):
    T = col_var.eltype
    sig = types.boolean(col_var, row_idx)

    target_info = TargetInfo()
    null_value = target_info.null_values[str(T)]
    # The server sends numbers as unsigned values rather than signed ones.
    # Thus, 129 should be read as -127 (overflow). See rbc issue #254
    nv = ir.Constant(ir.IntType(T.bitwidth), null_value)

    def codegen(context, builder, signature, args):
        ptr, index = args
        data = builder.extract_value(builder.load(ptr), [0])
        res = builder.load(builder.gep(data, [index]))

        if isinstance(T, types.Float):
            res = builder.bitcast(res, nv.type)

        return builder.icmp_signed('==', res, nv)

    return sig, codegen


# "BufferPointer.is_null" checks if a given array or column is null
# as opposed to "BufferType.is_null" that checks if an index in a
# column is null
@extending.overload_method(BufferPointer, 'is_null')
def omnisci_buffer_is_null(x, row_idx=None):
    if isinstance(x, BufferPointer):
        if cgutils.is_nonelike(row_idx):
            def impl(x, row_idx=None):
                return omnisci_buffer_is_null_(x)
        else:
            def impl(x, row_idx=None):
                return omnisci_buffer_idx_is_null_(x, row_idx)
        return impl


@extending.intrinsic
def omnisci_buffer_idx_set_null(typingctx, arr, row_idx):
    T = arr.eltype
    sig = types.none(arr, row_idx)

    target_info = TargetInfo()
    null_value = target_info.null_values[f'{T}']

    # The server sends numbers as unsigned values rather than signed ones.
    # Thus, 129 should be read as -127 (overflow). See rbc issue #254
    bitwidth = T.bitwidth
    null_value = np.dtype(f'uint{bitwidth}').type(null_value).view(f'int{bitwidth}')

    def codegen(context, builder, signature, args):
        # get the operator.setitem intrinsic
        fnop = context.typing_context.resolve_value_type(omnisci_buffer_ptr_setitem_)
        setitem_sig = types.none(arr, row_idx, T)
        # register the intrinsic in the typing ctx
        fnop.get_call_type(context.typing_context, setitem_sig.args, {})
        intrinsic = context.get_function(fnop, setitem_sig)

        data, index = args
        # data = {T*, i64, i8}*
        ty = data.type.pointee.elements[0].pointee
        nv = ir.Constant(ir.IntType(T.bitwidth), null_value)
        if isinstance(T, types.Float):
            nv = builder.bitcast(nv, ty)
        intrinsic(builder, (data, index, nv,))

    return sig, codegen


@extending.overload_method(BufferPointer, 'set_null')
def omnisci_buffer_set_null(x, row_idx=None):
    if isinstance(x, BufferPointer):
        if cgutils.is_nonelike(row_idx):
            def impl(x, row_idx=None):
                return omnisci_buffer_set_null_(x)
        else:
            def impl(x, row_idx=None):
                return omnisci_buffer_idx_set_null(x, row_idx)
            return impl
        return impl


@extending.overload_method(BufferPointer, 'free')
def omnisci_buffer_free(x):
    if isinstance(x, BufferPointer):
        def impl(x):
            return free_buffer(x)
        return impl


@extending.overload(operator.eq)
def dtype_eq(a, b):
    if isinstance(a, types.DTypeSpec) and isinstance(b, types.DTypeSpec):
        eq = (a == b)

        def impl(a, b):
            return eq
        return impl


@extending.overload_attribute(BufferPointer, 'dtype')
def omnisci_buffer_dtype(x):
    if isinstance(x, BufferPointer):
        dtype = x.eltype

        def impl(x):
            return dtype
        return impl


@extending.lower_builtin('iternext', BufferPointerIteratorType)
@imputils.iternext_impl(imputils.RefType.UNTRACKED)
def iternext_BufferPointer(context, builder, sig, args, result):
    [iterbufty] = sig.args
    [bufiter] = args

    iterval = context.make_helper(builder, iterbufty, value=bufiter)

    buf = iterval.buffer
    idx = builder.load(iterval.index)

    len_fn = context.typing_context.resolve_value_type(len)
    len_sig = types.intp(sig.args[0].buffer_type)
    # if the intrinsic was not called before, one need to "register" it first
    len_fn.get_call_type(context.typing_context, len_sig.args, {})
    count = context.get_function(len_fn, len_sig)(builder, [buf])

    is_valid = builder.icmp_signed('<', idx, count)
    result.set_valid(is_valid)

    with builder.if_then(is_valid):
        getitem_fn = context.typing_context.resolve_value_type(operator.getitem)
        getitem_sig = iterbufty.buffer_type.eltype(iterbufty.buffer_type, types.intp)
        # same here, "register" the intrinsic before calling it
        getitem_fn.get_call_type(context.typing_context, getitem_sig.args, {})
        getitem_out = context.get_function(getitem_fn, getitem_sig)(builder, [buf, idx])
        result.yield_(getitem_out)
        nidx = builder.add(idx, context.get_constant(types.intp, 1))
        builder.store(nidx, iterval.index)


@extending.lower_builtin('getiter', BufferPointer)
def getiter_buffer_pointer(context, builder, sig, args):
    [buffer] = args

    iterobj = context.make_helper(builder, sig.return_type)

    # set the index to zero
    zero = context.get_constant(types.uintp, 0)
    indexptr = cgutils.alloca_once_value(builder, zero)

    iterobj.index = indexptr

    # wire in the buffer type data
    iterobj.buffer = buffer

    res = iterobj._getvalue()
    return imputils.impl_ret_new_ref(context, builder, sig.return_type, res)



from rbc.utils import get_version
from numba.core import cgutils
from llvmlite import ir


int8_t = ir.IntType(8)
int32_t = ir.IntType(32)
int64_t = ir.IntType(64)
void_t = ir.VoidType()
fp32 = ir.FloatType()
fp64 = ir.DoubleType()


def get_or_insert_function(module, fnty, name=None):
    if get_version('numba') >= (0, 54):
        fn = cgutils.get_or_insert_function(
            module, fnty, name=name)
    else:
        fn = module.get_or_insert_function(
            fnty, name=name)
    return fn


def get_member_value(builder, data, idx):
    if get_version('numba') >= (0, 54):
        assert data.opname == 'insertvalue', data.opname
        return builder.extract_value(data, [idx])
    else:
        assert data.opname == 'load', data.opname
        struct = data.operands[0]
        return builder.load(builder.gep(struct, [int32_t(0), int32_t(idx)]))



import math
import pytest
import sys
import numpy as np
import numba as nb
import rbc.omniscidb as rbc_omnisci
from rbc.stdlib import array_api

available_version, reason = rbc_omnisci.is_available()
pytestmark = pytest.mark.skipif(not available_version, reason=reason)


@pytest.fixture(scope='module')
def nb_version():
    from rbc.utils import get_version
    return get_version('numba')


@pytest.fixture(scope='module')
def omnisci():
    # TODO: use omnisci_fixture from rbc/tests/__init__.py
    config = rbc_omnisci.get_client_config(debug=not True)
    m = rbc_omnisci.RemoteOmnisci(**config)
    table_name = 'rbc_test_omnisci_math'

    m.sql_execute(f'DROP TABLE IF EXISTS {table_name}')

    m.sql_execute(
        f'CREATE TABLE IF NOT EXISTS {table_name}'
        ' (a BOOLEAN, b BOOLEAN, x DOUBLE, y DOUBLE, z DOUBLE, i INT, '
        'j INT, t INT[], td DOUBLE[], te INT[]);')

    for _i in range(1, 6):
        a = str((_i % 3) == 0).lower()
        b = str((_i % 2) == 0).lower()
        x = 0.123 + _i/10.0
        y = _i/6.0
        z = _i + 1.23
        i = _i
        j = i * 10
        t = 'ARRAY[%s]' % (', '.join(str(j + i) for i in range(-i, i+1)))
        td = 'ARRAY[%s]' % (', '.join(str(j + i/1.0) for i in range(-i, i+1)))
        te = 'Array[]'
        m.sql_execute(
            f'insert into {table_name} values (\'{a}\', \'{b}\', {x}, {y},'
            f' {z}, {i}, {j}, {t}, {td}, {te})')

    m.table_name = table_name
    yield m

    m.sql_execute(f'DROP TABLE IF EXISTS {table_name}')


math_functions = [
    # Number-theoretic and representation functions
    ('ceil', 'int64(double)'),
    ('comb', 'int64(int64, int64)'),
    ('copysign', 'double(double, double)'),
    ('fabs', 'double(double)'),
    ('factorial', 'int64(int64)'),
    ('floor', 'double(double)'),
    ('fmod', 'double(double, double)'),
    ('frexp', 'double(double)'),  # returns a pair (m, e)
    ('fsum', 'double(double[])'),
    ('gcd', 'int(int, int)'),
    ('isclose', 'bool(double, double)'),
    ('isfinite', 'bool(double)'),
    ('isinf', 'bool(double)'),
    ('isnan', 'bool(double)'),
    ('isqrt', 'int64(int64)'),
    ('ldexp', 'double(double, int)'),
    ('modf', 'double(double, double)'),
    ('perm', 'int(int, int)'),
    ('prod', 'int64(int64[])'),
    ('remainder', 'double(double, double)'),
    ('trunc', 'double(double)'),
    # Power and logarithmic functions
    ('exp', 'double(double)'),
    ('expm1', 'double(double)'),
    ('log', 'double(double)'),
    ('log1p', 'double(double)'),
    ('log2', 'double(double)'),
    ('log10', 'double(double)'),
    ('pow', 'double(double, double)'),
    ('sqrt', 'double(double)'),
    # # Trigonometric functions
    ('acos', 'double(double)'),
    ('asin', 'double(double)'),
    ('atan', 'double(double)'),
    ('atan2', 'double(double, double)'),
    ('cos', 'double(double)'),
    ('dist', 'int64(int64[], int64[])'),
    ('hypot', 'double(double, double)'),
    ('sin', 'double(double)'),
    ('tan', 'double(double)'),
    ('degrees', 'double(double)'),
    ('radians', 'double(double)'),
    # # Hyperbolic functions
    ('acosh', 'double(double)'),
    ('asinh', 'double(double)'),
    ('atanh', 'double(double)'),
    ('cosh', 'double(double)'),
    ('sinh', 'double(double)'),
    ('tanh', 'double(double)'),
    # # Special functions
    ('erf', 'double(double)'),
    ('erfc', 'double(double)'),
    ('gamma', 'double(double)'),
    ('lgamma', 'double(double)'),
    # Constants
    ('pi', 'double(double)'),
    ('e', 'double(double)'),
    ('tau', 'double(double)'),
    ('inf', 'double(double)'),
    ('nan', 'double(double)'),
]


devices = ('cpu', 'gpu')


@pytest.mark.slow
@pytest.mark.parametrize("device", devices)
@pytest.mark.parametrize("fn_name, signature", math_functions,
                         ids=["math." + item[0] for item in math_functions])
def test_math_function(omnisci, device, nb_version, fn_name, signature):
    omnisci.reset()

    if not omnisci.has_cuda and device == 'gpu':
        pytest.skip('test requires CUDA-enabled omniscidb server')

    math_func = getattr(math, fn_name, None)
    if math_func is None:
        pytest.skip(f'{fn_name}: not available in {math.__name__} module'
                    f' of Python {sys.version.split(None, 1)[0]}')

    if fn_name in ['prod', 'comb', 'factorial', 'fsum',
                   'fmod', 'isclose', 'isqrt', 'modf', 'dist', 'perm']:
        pytest.skip(f'{fn_name}: Numba uses cpython implementation! [rbc issue 156]')

    if fn_name in ['frexp']:
        pytest.skip(f'{fn_name} returns a pair (m, e) [rbc issue 156/202]')

    arity = signature.count(',') + 1
    kind = signature.split('(')[1].split(',')[0].split(')')[0]

    if fn_name in ['pi', 'e', 'tau', 'inf', 'nan']:
        fn = eval(f'lambda x: math.{fn_name}', dict(math=math))
    elif arity == 1:
        fn = eval(f'lambda x: math.{fn_name}(x)', dict(math=math))
    elif arity == 2:
        fn = eval(f'lambda x, y: math.{fn_name}(x, y)',
                  dict(math=math))
    else:
        raise NotImplementedError((signature, arity))

    fprefix = 'rbc_test_'  # to avoid conflicts with SQL functions

    if fn.__name__ == '<lambda>':
        # give lambda function a name
        fn.__name__ = fn_name

    fn.__name__ = fprefix + fn.__name__

    omnisci(signature, devices=[device])(fn)

    omnisci.register()

    if kind == 'double':
        assert arity <= 3, arity
        xs = ', '.join('xyz'[:arity])
    elif kind.startswith('int'):
        assert arity <= 2, arity
        xs = ', '.join('ij'[:arity])
    elif kind.startswith('bool'):
        assert arity <= 2, arity
        xs = ', '.join('ab'[:arity])
    elif kind == 'constant':
        xs = ''
    else:
        raise NotImplementedError(kind)

    if fn_name in ['acosh', 'asinh']:
        xs = 'z'

    if fn_name in ['ldexp']:
        xs = 'x, i'

    query = f'select {xs}, {fprefix}{fn_name}({xs}) from {omnisci.table_name}'
    descr, result = omnisci.sql_execute(query)
    for args in list(result):
        result = args[-1]
        if fn_name in ['pi', 'e', 'tau', 'inf', 'nan']:
            expected = math_func
        else:
            expected = math_func(*args[:-1])
        if np.isnan(expected):
            assert np.isnan(result)
        else:
            assert(np.isclose(expected, result))


numpy_functions = [
    # Arithmetic functions:
    ('absolute', 'double(double)', np.absolute),
    ('conjugate', 'double(double)', np.conjugate),
    ('conj', 'double(double)', np.conjugate),
    ('fabs', 'double(double)', np.fabs),
    ('fmax', 'double(double, double)', np.fmax),
    ('fmin', 'double(double, double)', np.fmin),
    ('maximum', 'double(double, double)', np.maximum),
    ('minimum', 'double(double, double)', np.minimum),
    ('positive', 'double(double)', np.positive),
    ('negative', 'double(double)', np.negative),
    ('sign', 'double(double)', np.sign),
    ('reciprocal', 'double(double)', np.reciprocal),
    ('add', 'double(double, double)', np.add),
    ('subtract', 'double(double, double)', np.subtract),
    ('multiply', 'double(double, double)', np.multiply),
    ('divide', 'double(double, double)', np.divide),
    ('true_divide', 'double(double, double)', np.true_divide),
    ('floor_divide', 'double(double, double)', np.floor_divide),
    ('power', 'double(double, double)', np.power),
    ('float_power', 'double(double, double)', np.float_power),
    ('square', 'double(double)', np.square),
    ('sqrt', 'double(double)', np.sqrt),
    ('cbrt', 'double(double)', np.cbrt),   # not supported by numba
    ('remainder', 'double(double, double)', np.remainder),
    ('fmod', 'double(double, double)', np.fmod),
    ('modf', 'double(double, double)', np.mod),
    ('modi', 'int(int, int)', np.mod),
    ('divmod0', 'int(int, int)', lambda i, j: np.divmod(i, j)[0]),
    # Trigonometric functions:
    ('sin', 'double(double)', np.sin),
    ('cos', 'double(double)', np.cos),
    ('tan', 'double(double)', np.tan),
    ('arcsin', 'double(double)', np.arcsin),
    ('arccos', 'double(double)', np.arccos),
    ('arctan', 'double(double)', np.arctan),
    ('arctan2', 'double(double, double)', np.arctan2),
    ('hypot', 'double(double, double)', np.hypot),
    ('radians', 'double(double)', np.radians),
    ('rad2deg', 'double(double)', np.rad2deg),
    ('deg2rad', 'double(double)', np.deg2rad),
    ('degrees', 'double(double)', np.degrees),
    # Hyperbolic functions:
    ('sinh', 'double(double)', np.sinh),
    ('cosh', 'double(double)', np.cosh),
    ('tanh', 'double(double)', np.tanh),
    ('arcsinh', 'double(double)', np.arcsinh),
    ('arccosh', 'double(double)', np.arccosh),
    ('arctanh', 'double(double)', np.arctanh),
    # Exp-log functions:
    ('exp', 'double(double)', np.exp),
    ('expm1', 'double(double)', np.expm1),
    ('exp2', 'double(double)', np.exp2),
    ('log', 'double(double)', np.log),
    ('log10', 'double(double)', np.log10),
    ('log2', 'double(double)', np.log2),
    ('log1p', 'double(double)', np.log1p),
    ('logaddexp', 'double(double, double)', np.logaddexp),
    ('logaddexp2', 'double(double, double)', np.logaddexp2),
    ('ldexp', 'double(double, int)', np.ldexp),
    ('frexp0', 'double(double)', lambda x: np.frexp(x)[0]),
    # Rounding functions:
    ('around', 'double(double)', lambda x: np.around(x)),
    ('round2',  # round and round_ are not good names
     'double(double)', lambda x: np.round_(x)),  # force arity to 1
    ('floor', 'double(double)', np.floor),
    ('ceil', 'double(double)', np.ceil),
    ('trunc', 'double(double)', np.trunc),
    ('rint', 'double(double)', np.rint),
    ('spacing', 'double(double)', np.spacing),
    ('nextafter', 'double(double, double)', np.nextafter),
    # Rational functions:
    ('gcd', 'int(int, int)', np.gcd),
    ('lcm', 'int(int, int)', np.lcm),
    ('right_shift', 'int(int, int)', np.right_shift),
    ('left_shift', 'int64(int64, int64)', np.left_shift),
    # Misc functions
    ('heaviside', 'double(double, double)', np.heaviside),
    ('copysign', 'double(double, double)', np.copysign),
    # Bit functions:
    ('invert', 'int(int)', np.invert),
    ('bitwise_not', 'int(int)', np.bitwise_not),
    ('bitwise_or', 'int(int, int)', np.bitwise_or),
    ('bitwise_xor', 'int(int, int)', np.bitwise_xor),
    ('bitwise_and', 'int(int, int)', np.bitwise_and),
    # Logical functions:
    ('isfinite', 'bool(double)', np.isfinite),
    ('isinf', 'bool(double)', np.isinf),
    ('isnan', 'bool(double)', np.isnan),
    ('signbit', 'bool(double)', np.signbit),
    ('less', 'bool(double, double)', np.less),
    ('less_equal', 'bool(double, double)', np.less_equal),
    ('greater', 'bool(double, double)', np.greater),
    ('greater_equal', 'bool(double, double)', np.greater_equal),
    ('equal', 'bool(double, double)', np.equal),
    ('not_equal', 'bool(double, double)', np.not_equal),
    ('logical_or', 'bool(bool, bool)', np.logical_or),
    ('logical_xor', 'bool(bool, bool)', np.logical_xor),
    ('logical_and', 'bool(bool, bool)', np.logical_and),
    ('logical_not', 'bool(bool)', np.logical_not),
    # missing ufunc-s as unsupported: matmul, isnat
]

if np is not None:
    for n, f in np.__dict__.items():
        if n in ['matmul', 'isnat']:  # UNSUPPORTED
            continue
        if isinstance(f, np.ufunc):
            for item in numpy_functions:
                if item[0].startswith(f.__name__):
                    break
            else:
                print(f'TODO: ADD {n} TEST TO {__file__}')


@pytest.mark.slow
@pytest.mark.parametrize("device", devices)
@pytest.mark.parametrize("fn_name, signature, np_func", numpy_functions,
                         ids=["np." + item[0] for item in numpy_functions])
def test_numpy_function(omnisci, device, nb_version, fn_name, signature, np_func):
    omnisci.reset()

    if not omnisci.has_cuda and device == 'gpu':
        pytest.skip('Test requires CUDA-enabled omniscidb server')

    if fn_name in ['cbrt', 'float_power']:
        pytest.skip(f'Numba does not support {fn_name}')

    arity = signature.count(',') + 1
    kind = signature.split('(')[1].split(',')[0].split(')')[0]
    if isinstance(np_func, np.ufunc):
        # numba does not support jitting ufunc-s directly
        if arity == 1:
            fn = eval(f'lambda x: array_api.{np_func.__name__}(x)', dict(array_api=array_api))
        elif arity == 2:
            fn = eval(f'lambda x, y: array_api.{np_func.__name__}(x, y)',
                      dict(array_api=array_api))
        else:
            raise NotImplementedError((signature, arity))
    else:
        fn = np_func

    if fn.__name__ == '<lambda>':
        # give lambda function a name
        fn.__name__ = fn_name

    if fn_name in ['positive', 'divmod0', 'frexp0']:
        try:
            if arity == 1:
                nb.njit(fn)(0.5)
            elif arity == 2:
                nb.njit(fn)(0.5, 0.5)
        except nb.errors.TypingError as msg:
            msg = str(msg).splitlines()[1]
            pytest.skip(msg)

    if fn_name in ['spacing']:
        # Skipping spacing__cpu_0 that uses undefined function `npy_spacing`
        pytest.skip(f'{fn_name}: FIXME')

    if device == 'gpu' and fn_name in ['floor_divide', 'around', 'round2', 'round_']:
        pytest.skip(f'Missing libdevice bindigs for {fn_name}')

    omnisci(signature, devices=[device])(fn)

    omnisci.register()

    if fn_name == 'ldexp':
        xs = ', '.join('xi')
    elif fn_name in ['arccosh', 'arcsinh']:
        xs = 'z'
    elif kind == 'double':
        assert arity <= 3, arity
        xs = ', '.join('xyz'[:arity])
    elif kind.startswith('int'):
        assert arity <= 2, arity
        xs = ', '.join('ij'[:arity])
    elif kind.startswith('bool'):
        assert arity <= 2, arity
        xs = ', '.join('ab'[:arity])
    else:
        raise NotImplementedError(kind)

    query = f'select {xs}, {fn_name}({xs}) from {omnisci.table_name}'
    descr, result = omnisci.sql_execute(query)
    for args in list(result):
        result = args[-1]
        expected = np_func(*args[:-1])
        if np.isnan(expected):
            assert np.isnan(result), fn_name
        else:
            assert(np.isclose(expected, result)), fn_name
